from .download import download_zip, download_file
from pathlib import Path
down_path = Path(__file__).parent
xz_url = "addons_file"+"/"+down_path.name+"/"
download_file(xz_url + "ANM_Geo_Nodes.blend",down_path)
download_file(xz_url + "Curves_Presets.blend",down_path)

import bpy
from bpy.types import Panel, Operator
import os
from bpy_extras import view3d_utils
from mathutils import Vector 
import rna_keymap_ui
import bmesh

from .translation.translate import GetTranslationDict


bl_info = {
    "name": "Anime Hair Maker",
    "description": "A helper script to apply curves presets for anime hairs",
    "author": "Young <learnasimakeit@outlook.com>",
    "version": (1, 5, 4),
    "blender": (3, 5, 0),
    "location": "View3D > N > Anime Hair Maker",
    "category": "3D View"}



def Update_Bevel_Objects_Size(self, context):
    bevel_objects_size = context.scene.animehair_props.bevel_objects_size
    hair_cur = context.active_object

    if hair_cur.data.bevel_mode == 'OBJECT':
        cur_t = None
        for cur in hair_cur.children:
            if 'ANM_Taper' in cur.name:
                cur_t = cur
                break
        if cur_t:
            if len(cur_t.children):
                for c in cur_t.children:
                    c.scale[0] = bevel_objects_size
                    c.scale[1] = bevel_objects_size
                    c.scale[2] = bevel_objects_size
    elif hair_cur.data.bevel_mode == 'ROUND':
        hair_cur.data.bevel_depth = bevel_objects_size * 0.01


    return


def Update_geo_node_per_strand_tilt(self, context):
    geo_node_per_strand_tilt = context.scene.animehair_props.geo_node_per_strand_tilt
    hair_cur = context.active_object

    if 'ANM_Curl' in hair_cur.name or 'ANM_Braid' in hair_cur.name:
        if hair_cur.modifiers.get('ANM_Geo_Node'):
            hair_cur.modifiers['ANM_Geo_Node'].node_group.nodes['Transform'].inputs[2].default_value[2] = geo_node_per_strand_tilt

    return


    

def Update_Taper_Objects_Root(self, context):
    taper_objects_root = context.scene.animehair_props.taper_objects_root
    hair_cur = context.active_object
    taper_object = hair_cur.data.taper_object
    if taper_object is not None and taper_object.name[:5] == 'ANM_T':
        taper_object.data.splines[0].points[0].co.y = taper_objects_root

    return


def Update_Taper_Objects_Root_To_Mid(self, context):
    taper_objects_root_to_mid = context.scene.animehair_props.taper_objects_root_to_mid
    hair_cur = context.active_object
    taper_object = hair_cur.data.taper_object
    if taper_object is not None and taper_object.name[:5] == 'ANM_T':
        taper_object.data.splines[0].points[1].co.y = taper_objects_root_to_mid

    return


def Update_Taper_Objects_Mid(self, context):
    taper_objects_mid = context.scene.animehair_props.taper_objects_mid
    hair_cur = context.active_object
    taper_object = hair_cur.data.taper_object
    if taper_object is not None and taper_object.name[:5] == 'ANM_T':
        taper_object.data.splines[0].points[2].co.y = taper_objects_mid

    return


def Update_Taper_Objects_Mid_To_Tip(self, context):
    taper_objects_mid_to_tip = context.scene.animehair_props.taper_objects_mid_to_tip
    hair_cur = context.active_object
    taper_object = hair_cur.data.taper_object
    if taper_object is not None and taper_object.name[:5] == 'ANM_T':
        taper_object.data.splines[0].points[3].co.y = taper_objects_mid_to_tip

    return


def Update_Taper_Objects_Tip(self, context):
    taper_objects_tip = context.scene.animehair_props.taper_objects_tip
    hair_cur = context.active_object
    taper_object = hair_cur.data.taper_object
    if taper_object is not None and taper_object.name[:5] == 'ANM_T':
        taper_object.data.splines[0].points[4].co.y = taper_objects_tip

    return


def Update_Curve_Length_Resolution(self, context):
    curve_length_resolution = context.scene.animehair_props.curve_length_resolution
    hair_cur = context.active_object
    hair_cur.data.resolution_u = curve_length_resolution

    return


def Update_Curve_Bevel_Resolution(self, context):
    curve_bevel_resolution = context.scene.animehair_props.curve_bevel_resolution
    hair_cur = context.active_object
    if hair_cur.data.bevel_mode == 'OBJECT':
        bevel_cur = hair_cur.data.bevel_object
        if bevel_cur is not None:
            bevel_cur.data.resolution_u = curve_bevel_resolution
    elif hair_cur.data.bevel_mode == 'ROUND':
        hair_cur.data.bevel_resolution = curve_bevel_resolution

    return


def set_use_bezier(self, value):
    self["use_bezier"] = value

def Update_use_bezier(self, context):
    animehair_props = context.scene.animehair_props
    use_bezier = animehair_props.use_bezier

    if use_bezier:
        set_use_nurbs(animehair_props,False)
    else:
        set_use_nurbs(animehair_props,True)

    return

def set_use_nurbs(self, value):
    self["use_nurbs"] = value

def Update_use_nurbs(self, context):
    animehair_props = context.scene.animehair_props
    use_nurbs = animehair_props.use_nurbs

    if use_nurbs:
        set_use_bezier(animehair_props,False)
    else:
        set_use_bezier(animehair_props,True)

    return
   


def Update_geo_node_frequency(self, context):
    geo_node_frequency = context.scene.animehair_props.geo_node_frequency
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_2"] = geo_node_frequency
        hair_cur.update_tag()

    return

def Update_geo_node_length_res(self, context):
    geo_node_length_res = context.scene.animehair_props.geo_node_length_res
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_8"] = geo_node_length_res
        hair_cur.update_tag()

    return

def Update_geo_node_per_strand_res(self, context):
    geo_node_per_strand_res = context.scene.animehair_props.geo_node_per_strand_res
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_4"] = geo_node_per_strand_res
        hair_cur.update_tag()

    return


def Update_geo_node_per_strand_radius(self, context):
    geo_node_per_strand_radius = context.scene.animehair_props.geo_node_per_strand_radius
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_5"] = geo_node_per_strand_radius
        hair_cur.update_tag()

    return


def Update_geo_node_scale(self, context):
    geo_node_scale = context.scene.animehair_props.geo_node_scale
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_11"] = geo_node_scale
        hair_cur.update_tag()

    return

def Update_geo_node_stretch1(self, context):
    geo_node_stretch1 = context.scene.animehair_props.geo_node_stretch1
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_7"] = geo_node_stretch1
        hair_cur.update_tag()

    return

def Update_geo_node_stretch2(self, context):
    geo_node_stretch2 = context.scene.animehair_props.geo_node_stretch2
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_9"] = geo_node_stretch2
        hair_cur.update_tag()

    return

def Update_geo_node_subd(self, context):
    geo_node_subd = context.scene.animehair_props.geo_node_subd
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_10"] = geo_node_subd
        hair_cur.update_tag()

    return

def Update_geo_node_profile_res(self, context):
    geo_node_profile_res = context.scene.animehair_props.geo_node_profile_res
    hair_cur = context.active_object
    profile_cur = None
    if hair_cur.modifiers.get('ANM_Geo_Node'):
        profile_cur = hair_cur.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value

        if profile_cur:
            profile_cur.data.resolution_u = geo_node_profile_res

    return

def Update_fill_cap(self, context):
    fill_cap = context.scene.animehair_props.fill_cap
    hair_cur = context.active_object

    if hair_cur.type == 'CURVE':
        if hair_cur.modifiers.get('ANM_Geo_Node'):
            hair_cur.modifiers['ANM_Geo_Node'].node_group.nodes["Curve to Mesh"].inputs[2].default_value = fill_cap
        hair_cur.data.use_fill_caps = fill_cap        

    return


def Update_gn_haircurves_length_res(self, context):
    gn_haircurves_length_res = context.scene.animehair_props.gn_haircurves_length_res
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_8"] = gn_haircurves_length_res
        hair_cur.update_tag()

    return

def Update_gn_haircurves_strand_tilt(self, context):
    gn_haircurves_strand_tilt = context.scene.animehair_props.gn_haircurves_strand_tilt
    hair_cur = context.active_object

    if 'ANM_HairCurves' in hair_cur.name:
        if hair_cur.modifiers.get('ANM_Geo_Node'):
            hair_cur.modifiers['ANM_Geo_Node'].node_group.nodes['Transform'].inputs[2].default_value[2] = gn_haircurves_strand_tilt

    return

def Update_gn_haircurves_profile_res(self, context):
    gn_haircurves_profile_res = context.scene.animehair_props.gn_haircurves_profile_res
    hair_cur = context.active_object
    profile_cur = None
    if hair_cur.modifiers.get('ANM_Geo_Node'):
        profile_cur = hair_cur.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value
        if profile_cur:
            profile_cur.data.resolution_u = gn_haircurves_profile_res

    return


def Update_gn_haircurves_subd(self, context):
    gn_haircurves_subd = context.scene.animehair_props.gn_haircurves_subd
    hair_cur = context.active_object

    if hair_cur.modifiers.get('ANM_Geo_Node'):
        hair_cur.modifiers["ANM_Geo_Node"]["Input_10"] = gn_haircurves_subd
        hair_cur.update_tag()

    return


def Update_use_taper_object(self, context):
    use_taper_object = context.scene.animehair_props.use_taper_object
    hair_cur = context.active_object
    if hair_cur.type == 'CURVE':
        if use_taper_object and len(hair_cur.children):
            for child in hair_cur.children:
                if 'ANM_Taper'in child.name:
                    hair_cur.data.taper_object = child
                    break
        else:
            hair_cur.data.taper_object = None

    return



def Update_use_custom_bevel(self, context):
    use_custom_bevel = context.scene.animehair_props.use_custom_bevel
    hair_cur = context.active_object
    if hair_cur.type == 'CURVE':
        if use_custom_bevel:
            hair_cur.data.bevel_mode = 'OBJECT'
        else:
            hair_cur.data.bevel_mode = 'ROUND'

    return


class AnimeHairProperties(bpy.types.PropertyGroup):
    braids_strand_count: bpy.props.IntProperty(name='braids_strand_count', default=3, min=2, max=6)
    
    geo_node_per_strand_tilt:  bpy.props.FloatProperty(
        name='geo_node_per_strand_tilt', default=0, min=0, max=6.283, update=Update_geo_node_per_strand_tilt)
    geo_node_frequency:  bpy.props.FloatProperty(
        name='geo_node_frequency', default=10, min=1, max=100, update=Update_geo_node_frequency)
    geo_node_length_res:  bpy.props.FloatProperty(
        name='geo_node_length_res', default=0.006, min=0.001, max=1, update=Update_geo_node_length_res)
    geo_node_per_strand_res:  bpy.props.IntProperty(
        name='geo_node_per_strand_res', default=4, min=3, max=8, update=Update_geo_node_per_strand_res)
    geo_node_per_strand_radius:  bpy.props.FloatProperty(
        name='geo_node_per_strand_radius', default=0.006, min=0.001, max=1, update=Update_geo_node_per_strand_radius)
    geo_node_scale:  bpy.props.FloatProperty(
        name='geo_node_scale', default=0.05, min=0.01, max=10, update=Update_geo_node_scale)
    geo_node_stretch1:  bpy.props.FloatProperty(
        name='geo_node_stretch1', default=1, min=0.01, max=10, update=Update_geo_node_stretch1)
    geo_node_stretch2:  bpy.props.FloatProperty(
        name='geo_node_stretch2', default=1, min=0.01, max=10, update=Update_geo_node_stretch2)
    geo_node_subd:  bpy.props.IntProperty(
        name='geo_node_subd', default=1, min=0, max=4, update=Update_geo_node_subd)
    geo_node_profile_res:  bpy.props.IntProperty(
        name='geo_node_profile_res', default=3, min=1, soft_max=6, update=Update_geo_node_profile_res)


    gn_haircurves_strand_tilt:  bpy.props.FloatProperty(
        name='gn_haircurves_strand_tilt', default=0, min=0, max=6.283, update=Update_gn_haircurves_strand_tilt)
    gn_haircurves_length_res:  bpy.props.FloatProperty(
        name='gn_haircurves_length_res', default=0.006, min=0.001, max=1, update=Update_gn_haircurves_length_res)
    gn_haircurves_subd:  bpy.props.IntProperty(
        name='gn_haircurves_subd', default=1, min=0, max=4, update=Update_gn_haircurves_subd)
    gn_haircurves_profile_res:  bpy.props.IntProperty(
        name='gn_haircurves_profile_res', default=3, min=1, soft_max=6, update=Update_gn_haircurves_profile_res)


    root_vg: bpy.props.BoolProperty(name='root_vg', default=False)
    auto_merge_end: bpy.props.BoolProperty(name='auto_merge_end', default=False)
    fill_cap: bpy.props.BoolProperty(name='fill_cap', default=False,update=Update_fill_cap)

    curve_length_resolution: bpy.props.IntProperty(
        name='Curve Length Resolution', default=8, min=2, soft_max=15, update=Update_Curve_Length_Resolution)
    curve_bevel_resolution: bpy.props.IntProperty(
        name='Curve Bevel Resolution', default=3, min=1, soft_max=12, update=Update_Curve_Bevel_Resolution)

    bevel_objects_size:  bpy.props.FloatProperty(
        name='HairStrand Size', default=1, min=0.01, soft_max=10, update=Update_Bevel_Objects_Size)
    taper_objects_root:  bpy.props.FloatProperty(
        name='taper curve root', default=0.2, min=0, soft_max=3, update=Update_Taper_Objects_Root)
    taper_objects_root_to_mid:  bpy.props.FloatProperty(
        name='taper curve root to mid', default=0.35, min=0, soft_max=3, update=Update_Taper_Objects_Root_To_Mid)
    taper_objects_mid:  bpy.props.FloatProperty(
        name='taper curve mid', default=0.3, min=0, soft_max=3, update=Update_Taper_Objects_Mid)
    taper_objects_mid_to_tip:  bpy.props.FloatProperty(
        name='taper curve mid to tip', default=0.15, min=0, soft_max=3, update=Update_Taper_Objects_Mid_To_Tip)
    taper_objects_tip:  bpy.props.FloatProperty(
        name='taper curve tip', default=0, min=0, soft_max=3, update=Update_Taper_Objects_Tip)
    
    adjustment_tab: bpy.props.BoolProperty(name='adjustment_tab', default=False)

    color_tab: bpy.props.BoolProperty(name='color_tab', default=False)


    misc_tab: bpy.props.BoolProperty(name='misc_tab', default=False)

    bool_source_file_updated:  bpy.props.BoolProperty(name='bool_source_file_updated', default=False)

    use_bezier: bpy.props.BoolProperty(name='use_bezier', default=True , update=Update_use_bezier)

    use_nurbs: bpy.props.BoolProperty(name='use_bezier', default=False, update=Update_use_nurbs)

    use_taper_object: bpy.props.BoolProperty(name='use_taper_object', default=True, update=Update_use_taper_object)

    use_custom_bevel: bpy.props.BoolProperty(name='use_custom_bevel', default=True, update=Update_use_custom_bevel)


class ANIME_HAIR_MAKER:
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = "ANM"

class CURVE_TO_ANIME_HAIR_PT_main(ANIME_HAIR_MAKER,Panel):
    bl_idname = "CURVE_TO_ANIME_HAIR_PT_main"
    bl_label = "Anime Hair Maker"

    def draw(self, context):
        layout = self.layout
        animehair_props = context.scene.animehair_props
        obj = context.active_object

        if bpy.app.version[0]>=3 and bpy.app.version[1]>=3 or bpy.app.version[0]>=4:
            box = layout.box()
            row = box.row()
            row.scale_y = 1.5
            row.operator("anm_draw_bezier_hairs.operator", text='Draw Bezier Hair')


            row = box.row()
            row.scale_y = 1.5
            row.operator("anime_hair_with_new_hair_system.operator",text='Make Hair Curves')

            row = box.row()
            sub = row.row(align=True)
            sub.scale_x = 2
            sub.scale_y = 2
            sub.operator("anm_add_hair_on_target.operator",text='Plant Hair Strand')

            column = sub.column()
            sub.scale_x = 0.3
            column.scale_y = 0.5
            column.prop(animehair_props, 'use_bezier', text='Bezier',toggle=True)
            column.prop(animehair_props, 'use_nurbs', text='Nurbs',toggle=True)

            row = box.row()
            row.operator("anm_convert_particle_hair_to_curve.operator", text='Particle Hairs To Curves')

            
            layout.separator(factor = 0.3)
            box = layout.box()
            row = box.row()
            sub = row.row(align=True)
            sub.scale_x = 0.3
            sub.label(text='  ')
            sub.scale_x = 2.0
            sub.prop(animehair_props, 'color_tab', text='Color',toggle=True)
            sub.scale_x = 0.3
            sub.label(text='  ')

            if animehair_props.color_tab:
                if obj and obj.active_material and 'ANM_Mat' in obj.active_material.name:
                    color_ramp = obj.active_material.node_tree.nodes["ColorRamp"].color_ramp
                    row = box.row(align = True)
                    row.prop(color_ramp.elements[0], 'color',text = '')
                    row.prop(color_ramp.elements[0], 'position')
                    row = box.row(align = True)
                    row.prop(color_ramp.elements[1], 'color',text = '')
                    row.prop(color_ramp.elements[1], 'position')

                    row = box.row()
                    row.scale_y = 1.2
                    row.operator("anm_make_mat_unique.operator",text='Make Mat Unique')
                elif obj and not obj.active_material:
                    
                    row = box.row()
                    row.scale_y = 1.2
                    row.operator("anm_make_mat_unique.operator",text='Add Mat')


            layout.separator(factor = 0.3)
            box = layout.box()
            row = box.row()
            sub = row.row(align=True)
            sub.scale_x = 0.3
            sub.label(text='  ')
            sub.scale_x = 2.0
            sub.prop(animehair_props, 'adjustment_tab', text='Adjustments',toggle=True)
            sub.scale_x = 0.3
            sub.label(text='  ')

            if animehair_props.adjustment_tab:
                if obj:
                    row = box.row()
                    row.operator("simplifyhair.operator", text='Convert to Bezier')
                    row.operator("decimatecurvepoints.operator", text='Decimate')

                    row = box.row()
                    row.scale_y = 1.5
                    row.operator("make_anime_hair.operator",text='Simple Hair')

                    row.operator("make_curl_hair.operator",text='Curl Hair')
                    row = box.row()
                    row.scale_y = 1.5
                    sub = row.row(align=True)
                    sub.scale_x = 2.0
                    sub.operator("make_braids_hair.operator",text='Braids Hair')
                    sub.scale_x = 1
                    sub.prop(animehair_props, 'braids_strand_count',text='')

                    if "ANM_Simple_" in obj.name and obj.type == "CURVE":
                        
                        row = box.row()
                        row.scale_y = 1.3
                        row.operator("anm_reset_children_objects.operator", text='Make Unique')


                        row = box.row()
                        row.scale_y = 1.3
                        row.prop(animehair_props, 'use_custom_bevel', text='Use Custom Profile',toggle=True)


                        row = box.row()
                        row.scale_y = 1.5
                        row.operator("previous_hair_bevel.operator", text='Prev Hair')
                        row.operator("next_hair_bevel.operator", text='Next Hair')

                        row = box.row()
                        row.scale_y = 1.2
                        row.prop(animehair_props, 'curve_length_resolution',
                                text='Curve Res')

                        row.prop(animehair_props, 'curve_bevel_resolution',
                                text='Bevel Res')


                        box.separator(factor = 0.1)

                        row = box.row()
                        row.scale_y = 1.3
                        row.prop(animehair_props, 'bevel_objects_size',
                                    text='Overall Size',slider=True)
                        box.separator(factor = 0.1)
                            
                        
                        box.prop(animehair_props, 'use_taper_object',text='Use Taper Object', toggle=True)
                        if animehair_props.use_taper_object:
                            box.prop(animehair_props, 'taper_objects_root',
                                    text='Taper Shape Root', slider=True)
                            box.prop(animehair_props, 'taper_objects_root_to_mid',
                                    text='Taper Shape Root To Mid', slider=True)
                            box.prop(animehair_props, 'taper_objects_mid',
                                    text='Taper Shape Mid', slider=True)
                            box.prop(animehair_props, 'taper_objects_mid_to_tip',
                                    text='Taper Shape Mid to Tip', slider=True)
                            box.prop(animehair_props, 'taper_objects_tip',
                                    text='Taper Shape Tip', slider=True)

                        box.separator(factor = 0.1)
                        
                        box = layout.box()
                        row = box.row()
                        row.scale_y = 1.2
                        sub = row.row()
                        sub.scale_x = 1
                        sub.operator("anm_edit_hair_bevel_object.operator", text='Edit Bevel Curve Shape')
                        
                        row = box.row()
                        row.scale_y = 2
                        sub = row.row()
                        sub.scale_x = 2
                        
                        sub.operator("anm_finalize_hair.operator", text='Convert To Mesh')
                        sub.scale_x = 0.3
                        
                        sub.prop(context.space_data.overlay, 'show_wireframes',text='wire')

                        row = box.row(align = True)
                        row.prop(animehair_props, 'auto_merge_end',text='Auto Merge', toggle=True)
                        row.prop(animehair_props, 'fill_cap',text='Fill Cap', toggle=True)
                        row.prop(animehair_props, 'root_vg',text='Root VG', toggle=True)

                    elif "ANM_Bevel_" in obj.name and obj.type == "CURVE":
                        layout.operator("anm_hide_hair_bevel_objects.operator", text='Restore')
                    
                    elif "ANM_Curl_" in obj.name and obj.type == "CURVE":
                        if obj.modifiers.get('ANM_Geo_Node'):
                            geo_mod = obj.modifiers['ANM_Geo_Node']
                            node_group = geo_mod.node_group

                            box.prop(animehair_props, 'geo_node_frequency',text='Frequency')
                            box.prop(animehair_props, 'geo_node_length_res',text='Length Resolution')

                            if bpy.app.version[0]==4 and bpy.app.version[1]<1 or bpy.app.version[0] <4:
                                if not node_group.nodes["Switch"].inputs[1].default_value:
                                    box.prop(animehair_props, 'geo_node_per_strand_res',text='Per Strand Resolution')
                                    box.prop(animehair_props, 'geo_node_per_strand_radius',text='Per Strand Radius')
                            else:
                                if not node_group.nodes["Switch"].inputs[0].default_value:
                                    box.prop(animehair_props, 'geo_node_per_strand_res',text='Per Strand Resolution')
                                    box.prop(animehair_props, 'geo_node_per_strand_radius',text='Per Strand Radius')

                            box.prop(animehair_props, 'geo_node_scale',text='Scale')
                            box.prop(animehair_props, 'geo_node_stretch1',text='Stretch1')
                            box.prop(animehair_props, 'geo_node_stretch2',text='Stretch2')
                            box.prop(animehair_props, 'geo_node_subd',text='Subdivision')


                            row = box.row()
                            if bpy.app.version[0]==4 and bpy.app.version[1]<1 or bpy.app.version[0] <4:
                                row.prop(node_group.nodes["Switch"].inputs[1], 'default_value',text='Use Custom Profile',toggle = True)
                                if node_group.nodes["Switch"].inputs[1].default_value:
                                    row = box.row()
                                    row.scale_y = 1.3
                                    row.operator("anm_reset_children_objects.operator", text='Make Unique')
                                    row = box.row()
                                    row.scale_y = 1.5
                                    row.operator("previous_hair_bevel.operator", text='Prev')
                                    row.operator("next_hair_bevel.operator", text='Next')

                                    row = box.row()
                                    row.scale_y = 1.2
                                    row.prop(animehair_props, 'geo_node_profile_res',text='Custom Profile Res')

                                    box.prop(animehair_props, 'geo_node_per_strand_tilt',text='Per Strand Tilt')
                                    box.prop(node_group.nodes["Value"].outputs[0], 'default_value',text='Per Strand Radius')


                                    box.operator("anm_edit_hair_geo_node_profile_object.operator", text='Edit Profile Curve Shape')
                            else:
                                row.prop(node_group.nodes["Switch"].inputs[0], 'default_value',text='Use Custom Profile',toggle = True)
                                if node_group.nodes["Switch"].inputs[0].default_value:
                                    row = box.row()
                                    row.scale_y = 1.3
                                    row.operator("anm_reset_children_objects.operator", text='Make Unique')
                                    row = box.row()
                                    row.scale_y = 1.5
                                    row.operator("previous_hair_bevel.operator", text='Prev')
                                    row.operator("next_hair_bevel.operator", text='Next')

                                    row = box.row()
                                    row.scale_y = 1.2
                                    row.prop(animehair_props, 'geo_node_profile_res',text='Custom Profile Res')

                                    box.prop(animehair_props, 'geo_node_per_strand_tilt',text='Per Strand Tilt')
                                    box.prop(node_group.nodes["Value"].outputs[0], 'default_value',text='Per Strand Radius')


                                    box.operator("anm_edit_hair_geo_node_profile_object.operator", text='Edit Profile Curve Shape')


                            box.separator(factor = 0.3)
                            row = box.row()
                            row.scale_y = 1.5
                            row.operator("anm_convert_curl_braids_to_simple_hair.operator", text='Convert To Simple Hair')


                            row = box.row()
                            row.scale_y = 2
                            sub = row.row()
                            sub.scale_x = 2
                            
                            sub.operator("anm_finalize_curl_braids_hair.operator", text='Convert To Mesh')
                            sub.scale_x = 0.3
                            
                            sub.prop(context.space_data.overlay, 'show_wireframes',text='wire')
                            
                            row = box.row(align = True)
                            row.prop(animehair_props, 'auto_merge_end',text='Auto Merge', toggle=True)
                            row.prop(animehair_props, 'fill_cap',text='Fill Cap', toggle=True)

                    elif "ANM_Braid" in obj.name and obj.type == "CURVE":
                        if obj.modifiers.get('ANM_Geo_Node'):
                            geo_mod = obj.modifiers['ANM_Geo_Node']
                            node_group = geo_mod.node_group

                            box.prop(animehair_props, 'geo_node_frequency',text='Frequency')
                            box.prop(animehair_props, 'geo_node_length_res',text='Length Resolution')
                            if bpy.app.version[0]==4 and bpy.app.version[1]<1 or bpy.app.version[0] <4:
                                if not node_group.nodes["Switch"].inputs[1].default_value:
                                    box.prop(animehair_props, 'geo_node_per_strand_res',text='Per Strand Resolution')
                                    box.prop(animehair_props, 'geo_node_per_strand_radius',text='Per Strand Radius')
                            else:
                                if not node_group.nodes["Switch"].inputs[0].default_value:
                                    box.prop(animehair_props, 'geo_node_per_strand_res',text='Per Strand Resolution')
                                    box.prop(animehair_props, 'geo_node_per_strand_radius',text='Per Strand Radius')
                            box.prop(animehair_props, 'geo_node_scale',text='Scale')
                            box.prop(animehair_props, 'geo_node_stretch1',text='Stretch1')
                            box.prop(animehair_props, 'geo_node_stretch2',text='Stretch2')
                            box.prop(animehair_props, 'geo_node_subd',text='Subdivision')


                            row = box.row()
                            if bpy.app.version[0]==4 and bpy.app.version[1]<1 or bpy.app.version[0] <4:
                                row.prop(node_group.nodes["Switch"].inputs[1], 'default_value',text='Use Custom Profile',toggle = True)
                                if node_group.nodes["Switch"].inputs[1].default_value:
                                    row = box.row()
                                    row.scale_y = 1.3
                                    row.operator("anm_reset_children_objects.operator", text='Make Unique')
                                    row = box.row()
                                    row.scale_y = 1.5
                                    row.operator("previous_hair_bevel.operator", text='Prev')
                                    row.operator("next_hair_bevel.operator", text='Next')

                                    row = box.row()
                                    row.scale_y = 1.2

                                    row.prop(animehair_props, 'geo_node_profile_res',text='Custom Profile Res')

                                    box.prop(animehair_props, 'geo_node_per_strand_tilt',text='Per Strand Tilt')
                                    box.prop(node_group.nodes["Value"].outputs[0], 'default_value',text='Per Strand Radius')

                                    box.operator("anm_edit_hair_geo_node_profile_object.operator", text='Edit Profile Curve Shape')
                            else:
                                row.prop(node_group.nodes["Switch"].inputs[0], 'default_value',text='Use Custom Profile',toggle = True)
                                if node_group.nodes["Switch"].inputs[0].default_value:
                                    row = box.row()
                                    row.scale_y = 1.3
                                    row.operator("anm_reset_children_objects.operator", text='Make Unique')
                                    row = box.row()
                                    row.scale_y = 1.5
                                    row.operator("previous_hair_bevel.operator", text='Prev')
                                    row.operator("next_hair_bevel.operator", text='Next')

                                    row = box.row()
                                    row.scale_y = 1.2

                                    row.prop(animehair_props, 'geo_node_profile_res',text='Custom Profile Res')

                                    box.prop(animehair_props, 'geo_node_per_strand_tilt',text='Per Strand Tilt')
                                    box.prop(node_group.nodes["Value"].outputs[0], 'default_value',text='Per Strand Radius')

                                    box.operator("anm_edit_hair_geo_node_profile_object.operator", text='Edit Profile Curve Shape')

                        

                            box.separator(factor = 0.3)
                            row = box.row()
                            row.scale_y = 1.5
                            row.operator("anm_convert_curl_braids_to_simple_hair.operator", text='Convert To Simple Hair')

                            row = box.row()
                            row.scale_y = 2
                            sub = row.row()
                            sub.scale_x = 2
                            
                            sub.operator("anm_finalize_curl_braids_hair.operator", text='Convert To Mesh')
                            sub.scale_x = 0.3
                            
                            sub.prop(context.space_data.overlay, 'show_wireframes',text='wire')

                            row = box.row(align = True)
                            row.prop(animehair_props, 'auto_merge_end',text='Auto Merge', toggle=True)
                            row.prop(animehair_props, 'fill_cap',text='Fill Cap', toggle=True)

                            
                    elif "ANM_HairCurves" in obj.name:
                        if obj.modifiers.get('ANM_Geo_Node'):
                            geo_mod = obj.modifiers['ANM_Geo_Node']
                            node_group = geo_mod.node_group

                            row = box.row()
                            row.scale_y = 1.3
                            row.operator("anm_reset_children_objects.operator", text='Make Unique')

                            row = box.row()
                            row.scale_y = 1.5
                            row.operator("previous_hair_bevel.operator", text='Prev')
                            row.operator("next_hair_bevel.operator", text='Next')

                            box.prop(animehair_props, 'gn_haircurves_length_res',text='Length Resolution')
                            box.prop(animehair_props, 'gn_haircurves_profile_res',text='Profile Resolution')
                            box.prop(node_group.nodes["Value"].outputs[0], 'default_value',text='Strand Radius')
                            box.prop(animehair_props, 'gn_haircurves_strand_tilt',text='Strand Tilt')
                            box.prop(animehair_props, 'gn_haircurves_subd',text='Subdivision')

                            row = box.row()
                            row.scale_y = 1.3
                            sub = row.row()
                            sub.scale_x = 2
                            sub.operator("anm_edit_hair_geo_node_profile_object.operator", text='Edit Profile Shape')
                            sub.scale_x = 0.7
                            sub.prop(context.space_data.overlay, 'show_wireframes',text='wire')
                            row = box.row()
                            row.scale_y = 2
                            if obj.type == 'CURVES':
                                row.operator("anm_curves_to_simple_curve.operator", text='Convert To Curve')
                                
                            elif obj.type == 'CURVE':
                                row.operator("anm_curves_to_mesh.operator", text='Convert To Mesh')
                                row = box.row(align = True)
                                row.prop(animehair_props, 'auto_merge_end',text='Auto Merge', toggle=True)
                                row.prop(animehair_props, 'fill_cap',text='Fill Cap', toggle=True)
                            


            layout.separator(factor = 0.3)
            box = layout.box()
            row = box.row()
            sub = row.row(align=True)
            sub.scale_x = 0.3
            sub.label(text='  ')
            sub.scale_x = 2.0
            sub.prop(animehair_props, 'misc_tab', text='MISC',toggle=True)
            sub.scale_x = 0.3
            sub.label(text='  ')


            if animehair_props.misc_tab:
                

                row = box.row()
                row.scale_y = 1.5
                row.operator("reset_hair_curve_tilt.operator", text='Reset Tilt')

                row.scale_y = 1.5
                row.operator("reset_hair_curve_radius.operator", text='Reset Radius')


                row = box.row()
                row.scale_y = 1.5
                row.operator("anm_create_bones_from_curves.operator", text='Create Bones From Curve Point')
                
                row = box.row()
                row.scale_y = 1.2
                row.operator("anm_move_curve_point.operator", text='Move Curve Point')

                row = box.row()
                row.scale_y = 1.2
                row.operator("anm_separate_to_new_curve.operator", text='Separate Selected To Curves')
                row = box.row()
                row.scale_y = 1.2
                row.operator("anm_isolate_selected_strand_show_all.operator", text='Isolate Selected /Show All Curves') 
                row = box.row()
                row.scale_y = 1.2
                row.operator("anm_hide_selected_strand.operator", text='Hide Selected Curves')
                row = box.row()
                row.scale_y = 1.2
                row.operator("anm_selected_prev_point.operator", text='Prev Curve Point')
                row.operator("anm_selected_next_point.operator", text='Next Curve Point')
                row = box.row()
                row.scale_y = 1.2
                row.operator("anm_select_left_handle.operator", text='Left Handle')
                row.operator("anm_select_main_handle.operator", text='Main Point')
                row.operator("anm_select_right_handle.operator", text='Right Handle')

                row = box.row()
                row.scale_y = 1.5
                row.operator("add_curve_to_source_file.operator", text='Add Curve To Source')
              
        else:
            layout.label(text = 'Please Use Blender 3.3 or Above')

def error_happened(self, context):
    self.layout.label(text="Some went wrong in file path or Blender file, Please double check.")



def Import_curves_presets():
    pref_path = bpy.context.preferences.addons[__package__].preferences.filepath
    directory = ''
    if pref_path == '':
        script_file = os.path.realpath(__file__)
        directory = os.path.dirname(script_file)
    else:
        directory = pref_path

    filepath = os.path.join(directory , "Curves_Presets.blend")  
    coll_name = "ANM_Curves_profile"


    try:
        if not bpy.data.scenes.get("ANM_Curves_profile"):
            bpy.data.scenes.new("ANM_Curves_profile")

        with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
            data_to.collections = [
                c for c in data_from.collections if c.startswith(coll_name)]

        for coll in data_to.collections:
            if coll:
                bpy.data.scenes["ANM_Curves_profile"].collection.children.link(coll)
    except:
        bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')

    return




def Import_ANM_Mat():
    if not bpy.data.materials.get('ANM_Mat'):
        script_file = os.path.realpath(__file__)
        directory = os.path.dirname(script_file)
        filepath = os.path.join(directory , "ANM_Geo_Nodes.blend") 
        mat_name = 'ANM_Mat'
        if not bpy.data.materials.get(mat_name):
            with bpy.data.libraries.load(filepath) as (data_from, data_to):
                data_to.materials = data_from.materials

    return

    




def Apply_existing_taper_value(current_curve):
    animehair_props =  bpy.context.scene.animehair_props
    
    cur_t = None
    for cur in current_curve.children:
        if 'ANM_Taper' in cur.name:
            cur_t = cur
            break
    if cur_t:
        if len(cur_t.children):
            for c in cur_t.children:
                c.scale[0] = animehair_props.bevel_objects_size
                c.scale[1] = animehair_props.bevel_objects_size
                c.scale[2] = animehair_props.bevel_objects_size
    if current_curve.data.taper_object:
        current_curve.data.taper_object.data.splines[0].points[0].co.y = animehair_props.taper_objects_root
        current_curve.data.taper_object.data.splines[0].points[1].co.y = animehair_props.taper_objects_root_to_mid
        current_curve.data.taper_object.data.splines[0].points[2].co.y = animehair_props.taper_objects_mid
        current_curve.data.taper_object.data.splines[0].points[3].co.y = animehair_props.taper_objects_mid_to_tip
        current_curve.data.taper_object.data.splines[0].points[4].co.y = animehair_props.taper_objects_tip
    if current_curve.data.bevel_object:
        current_curve.data.bevel_object.data.resolution_u = animehair_props.curve_bevel_resolution
    
    current_curve.data.resolution_u = animehair_props.curve_length_resolution
    

    return


class Convert_Particle_Hair_To_Curve_OP (Operator):
    """Convert Active Legacy Particle Hair System to Curve"""
    bl_label = "Convert Particle Hair To Curve"
    bl_idname = "anm_convert_particle_hair_to_curve.operator"

    @ classmethod
    def poll(cls, context):
        c = False
        if context.active_object != None:
            c = True
        return c

    def execute(self, context):
        obj = context.active_object
        if not obj.select_get():
            obj.select_set(True)

        if obj.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        for m in obj.modifiers:
            if m.type == "PARTICLE_SYSTEM" and m.show_viewport:
                bpy.context.view_layer.objects.active = obj
                bpy.ops.object.modifier_convert(modifier=m.name)
                bpy.ops.object.convert(target='CURVE')
                bpy.ops.object.origin_set(
                    type='ORIGIN_GEOMETRY', center='MEDIAN')
                context.active_object.name = 'ANM_' + m.particle_system.name

        return {'FINISHED'}

class Separate_To_New_Curve_OP (Operator):
    """Separate Selected Strand to New Curve"""
    bl_label = "Separate Selected Spline to New Curve"
    bl_idname = "anm_separate_to_new_curve.operator"

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True
        return c


    def execute(self, context):
        obj = context.active_object
        if not obj.select_get():
            obj.select_set(True)

        bpy.ops.curve.select_linked()
        bpy.ops.curve.separate()

        return {'FINISHED'}


class Previous_Hair_OP(Operator):
    """Load Previous Hair Bevel Profile for Hair Curves"""
    bl_label = "Previous Hair"
    bl_idname = "previous_hair_bevel.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if obj == objs[0]:
                if obj.type == 'CURVE' or obj.type == 'CURVES':
                    c = True
        return c

    def execute(self, context):
        obj = context.active_object

        if obj.type == 'CURVE':
            if not obj.modifiers.get('ANM_Geo_Node'):
                bevel_object= obj.data.bevel_object
                cur_t = None
                if len(obj.children):
                    for ob in obj.children:
                        if 'ANM_Taper' in ob.name:
                            cur_t = ob
                            break
                if cur_t and len(cur_t.children):

                    if bevel_object not in cur_t.children:
                        obj.data.bevel_object = cur_t.children[0]
                        bevel_object = cur_t.children[0]

                    indx = 0
                    for i, ob in enumerate(cur_t.children):
                        if ob.name == bevel_object.name:
                            indx = i
                            break

                    if indx-1 <  0 :
                        indx = len(cur_t.children)-1
                    else:
                        indx = indx-1

                    obj.data.bevel_object = cur_t.children[indx]

            elif obj.modifiers.get('ANM_Geo_Node'):
                profile_cur = obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value
                
                cur_t = None
                if len(obj.children):
                    for ob in obj.children:
                        if 'ANM_Taper' in ob.name:
                            cur_t = ob
                            break

                if cur_t and len(cur_t.children):
                    if profile_cur not in cur_t.children:
                        obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]
                    
                    indx = 0
                    for i, ob in enumerate(cur_t.children):
                        if ob.name == profile_cur.name:
                            indx = i
                            break   
                    if indx-1 <  0 :
                        indx = len(cur_t.children)-1
                    else:
                        indx = indx-1
                    obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[indx]
        
        elif obj.type == 'CURVES':
            if obj.modifiers.get('ANM_Geo_Node'):
                profile_cur = obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value
                
                cur_t = None
                if len(obj.children):
                    for ob in obj.children:
                        if 'ANM_Taper' in ob.name:
                            cur_t = ob
                            break

                if cur_t and len(cur_t.children):
                    if profile_cur not in cur_t.children:
                        obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]
                    
                    indx = 0
                    for i, ob in enumerate(cur_t.children):
                        if ob.name == profile_cur.name:
                            indx = i
                            break
                    
                    if indx-1 <  0 :
                        indx = len(cur_t.children)-1
                    else:
                        indx = indx-1
                    obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[indx]

        return {'FINISHED'}

class Next_Hair_OP(Operator):
    """Load Next Hair Bevel Profile for Hair Curves"""
    bl_label = "Next Hair"
    bl_idname = "next_hair_bevel.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if obj == objs[0]:
                if obj.type == 'CURVE' or obj.type == 'CURVES':
                    c = True
        return c

    def execute(self, context):
        obj = context.active_object
        if obj.type == 'CURVE':
            if not obj.modifiers.get('ANM_Geo_Node'):
                bevel_object= obj.data.bevel_object
                cur_t = None
                if len(obj.children):
                    for ob in obj.children:
                        if 'ANM_Taper' in ob.name:
                            cur_t = ob
                            break

                if cur_t and len(cur_t.children):

                    if bevel_object not in cur_t.children:
                        obj.data.bevel_object = cur_t.children[0]
                        bevel_object = cur_t.children[0]

                    indx = cur_t.children.index(bevel_object)
                    if indx+1 >  len(cur_t.children) -1 :
                        indx = 0
                    else:
                        indx = indx+1

                    obj.data.bevel_object = cur_t.children[indx]

            elif obj.modifiers.get('ANM_Geo_Node'):
                profile_cur = obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value
                
                cur_t = None
                if len(obj.children):
                    for ob in obj.children:
                        if 'ANM_Taper' in ob.name:
                            cur_t = ob
                            break
                
                if cur_t and len(cur_t.children):
                    if profile_cur not in cur_t.children:
                        obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]
                    
                    indx = 0
                    for i, ob in enumerate(cur_t.children):
                        if ob.name == profile_cur.name:
                            indx = i
                            break

                    if indx+1 >  len(cur_t.children) -1 :
                        indx = 0
                    else:
                        indx = indx+1
                    obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[indx]

        elif obj.type == 'CURVES':
            if obj.modifiers.get('ANM_Geo_Node'):
                profile_cur = obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value
                
                cur_t = None
                if len(obj.children):
                    for ob in obj.children:
                        if 'ANM_Taper' in ob.name:
                            cur_t = ob
                            break
                
                if cur_t and len(cur_t.children):
                    if profile_cur not in cur_t.children:
                        obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]
                    
                    indx = 0
                    for i, ob in enumerate(cur_t.children):
                        if ob.name == profile_cur.name:
                            indx = i
                            break
                    if indx+1 >  len(cur_t.children) -1 :
                        indx = 0
                    else:
                        indx = indx+1
                    obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[indx]

        return {'FINISHED'}


class Decimate_Curve_Points_OP(Operator):
    """decimate operator"""
    bl_label = "Decimate"
    bl_idname = "decimatecurvepoints.operator"
    bl_options = {'REGISTER', 'UNDO'}
    decimate_percentage: bpy.props.FloatProperty(default=0.35, min=0.001, max=1)

    @ classmethod
    def poll(cls, context):
        c = False
        if context.active_object != None:
            if context.active_object.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object
        if not hair_cur.select_get():
            hair_cur.select_set(True)

        if hair_cur.mode != "EDIT":
            bpy.ops.object.mode_set(mode='EDIT')

        bpy.ops.curve.select_all(action='SELECT')
        bpy.ops.curve.decimate(ratio=self.decimate_percentage)
        return {'FINISHED'}

def Move_Object_to_Hair_Curves_Collection(my_object):

    if my_object.type == 'CURVES':
        if not bpy.data.collections.get('ANM_Curves'):
            hair_curves_coll = bpy.data.collections.new('ANM_Curves')
            bpy.context.scene.collection.children.link(hair_curves_coll)
        my_collection = bpy.data.collections.get('ANM_Curves')

        for coll in my_object.users_collection:
            coll.objects.unlink(my_object)
        my_collection.objects.link(my_object)
    
    elif my_object.type == 'CURVE':
        if not bpy.data.collections.get('ANM_Simple_Curve'):
            hair_curves_coll = bpy.data.collections.new('ANM_Simple_Curve')
            bpy.context.scene.collection.children.link(hair_curves_coll)
        my_collection = bpy.data.collections.get('ANM_Simple_Curve')

        for coll in my_object.users_collection:
            coll.objects.unlink(my_object)
        my_collection.objects.link(my_object)

    elif my_object.type == 'MESH':
        if not bpy.data.collections.get('ANM_Mesh'):
            hair_curves_coll = bpy.data.collections.new('ANM_Mesh')
            bpy.context.scene.collection.children.link(hair_curves_coll)
        my_collection = bpy.data.collections.get('ANM_Mesh')

        for coll in my_object.users_collection:
            coll.objects.unlink(my_object)
        my_collection.objects.link(my_object)

    return


class Make_Anime_Hair_OP(Operator):
    """Apply preset to active curve object"""
    bl_label = "Make Anime Hair From Curve"
    bl_idname = "make_anime_hair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0] == obj and obj.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object
        org_indx = 0
        if hair_cur.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)


        if not hair_cur.active_material:
            Import_ANM_Mat()
            mat = bpy.data.materials['ANM_Mat']
            hair_cur.data.materials.append(mat)


        if "ANM_Curl_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Curl_","ANM_Simple_")
        elif "ANM_Braid2_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid2_","ANM_Simple_")
        elif "ANM_Braid3_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid3_","ANM_Simple_")
        elif "ANM_Braid4_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid4_","ANM_Simple_")
        elif "ANM_Braid5_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid5_","ANM_Simple_")
        elif "ANM_Braid6_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid6_","ANM_Simple_")
        elif "ANM_HairCurves_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_HairCurves_","ANM_Simple_")
            org_cur_t = None
            for obj in hair_cur.children:
                if "ANM_Taper" in obj.name:
                    org_cur_t = obj
            if org_cur_t and len(org_cur_t.children) and hair_cur.modifiers.get('ANM_Geo_Node'):
                org_indx = org_cur_t.children.index(hair_cur.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value)
                

        elif not "ANM_Simple_" in hair_cur.name:
            hair_cur.name = "ANM_Simple_" + hair_cur.name

        animehair_props =  context.scene.animehair_props
        
        try:
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()

            elif bpy.data.collections.get('ANM_Curves_profile') and animehair_props.bool_source_file_updated:
                ANM_Curves_collection = bpy.data.collections.get('ANM_Curves_profile')
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()
                
                Import_curves_presets()

                animehair_props.bool_source_file_updated = False

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()
        
            
            for obj in hair_cur.children:
                if "ANM_Taper" in obj.name:
                    if len(obj.children):
                        for ob in obj.children:
                            bpy.data.objects.remove(ob, do_unlink = True)
                    bpy.data.objects.remove(obj, do_unlink = True)

            for cur in bpy.data.curves:
                if not cur.users:
                    bpy.data.curves.remove(cur)


            for mod in hair_cur.modifiers:
                if mod.type == 'NODES':
                    if 'ANM' in mod.name :
                        hair_cur.modifiers.remove(mod)

            cur_coll = bpy.data.collections['ANM_Curves_profile']
            cur_coll.hide_viewport = True
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']


            Move_Object_to_Hair_Curves_Collection(hair_cur)
            hair_cur_coll = hair_cur.users_collection[0]
            bpy.ops.object.shade_smooth()
            
            hair_cur.data.bevel_mode = 'OBJECT' 
            hair_cur.data.bevel_depth = 0

            cur_t = None
            for cur in cur_coll.objects:
                if cur.name[:5] == 'ANM_T':
                    cur_t = bpy.data.objects.new(cur.name, cur.data)
                    cur_t.data = cur_t.data.copy()
                    hair_cur_coll.objects.link(cur_t)
                    hair_cur.data.use_path = False
                    cur_t.parent = hair_cur
                    cur_t.hide_viewport = True
                    cur_t.hide_render = True
            for cur in bevel_coll.objects:
                if cur.type  == 'CURVE':
                    cur_b = bpy.data.objects.new(cur.name, cur.data)
                    cur_b.data = cur_b.data.copy()
                    
                    hair_cur_coll.objects.link(cur_b)
                    cur_t.data.use_path = False
                    cur_b.parent = cur_t

                    cur_b.hide_viewport = True
                    cur_b.hide_render = True

            hair_cur.data.taper_object = cur_t
            hair_cur.data.bevel_object = cur_t.children[org_indx]

            bpy.context.view_layer.objects.active = hair_cur

            Apply_existing_taper_value(hair_cur)
            bpy.ops.outliner.orphans_purge()
        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')



        return {'FINISHED'}


class Make_Braids_Hair_OP(Operator):
    """Make Braids Hair, strands counts based on user selection"""
    bl_label = "Make Braids Hair From Curve"
    bl_idname = "make_braids_hair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0] == obj and obj.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object
        animehair_props = context.scene.animehair_props
        braids_strand_count =  context.scene.animehair_props.braids_strand_count
        
        if hair_cur.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)


        braid_name = 'ANM_Braid' + str(braids_strand_count) + '_'

        if "ANM_Curl_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Curl_",braid_name)
        elif "ANM_Simple_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Simple_",braid_name)
        elif "ANM_Braid2_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid2_",braid_name)
        elif "ANM_Braid3_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid3_",braid_name)
        elif "ANM_Braid4_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid4_",braid_name)
        elif "ANM_Braid5_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid5_",braid_name)
        elif "ANM_Braid6_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid6_",braid_name)
        elif "ANM_HairCurves_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_HairCurves_",braid_name)
        elif not braid_name in hair_cur.name:
            hair_cur.name = braid_name + hair_cur.name


        for mod in hair_cur.modifiers:
            if mod.type == 'NODES':
                if 'ANM_' in mod.name:
                    hair_cur.modifiers.remove(mod)
        
        try:
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()

            elif bpy.data.collections.get('ANM_Curves_profile') and animehair_props.bool_source_file_updated:
                ANM_Curves_collection = bpy.data.collections.get('ANM_Curves_profile')
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()
                
                Import_curves_presets()

                animehair_props.bool_source_file_updated = False

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()


            script_file = os.path.realpath(__file__)
            directory = os.path.dirname(script_file)
            
            
            nod_gp_name = 'ANM_Braid' + str(braids_strand_count) +'_GN'
            filepath = os.path.join(directory , "ANM_Geo_Nodes.blend")  
            if not bpy.data.node_groups.get(nod_gp_name):
                with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                    data_to.node_groups = [
                        name for name in data_from.node_groups if name.startswith(nod_gp_name)]

            bpy.data.node_groups[nod_gp_name].use_fake_user = True

            geo_node = hair_cur.modifiers.new(nod_gp_name,'NODES')
            geo_node.name = "ANM_Geo_Node"
            geo_node.node_group = bpy.data.node_groups[nod_gp_name]
            geo_node.node_group = geo_node.node_group.copy()
            
        
            for obj in hair_cur.children:
                if "ANM_Taper" in obj.name:
                    if len(obj.children):
                        for ob in obj.children:
                            bpy.data.objects.remove(ob, do_unlink = True)
                    bpy.data.objects.remove(obj, do_unlink = True)
                

            for cur in bpy.data.curves:
                if not cur.users:
                    bpy.data.curves.remove(cur)

            cur_coll = bpy.data.collections['ANM_Curves_profile']
            cur_coll.hide_viewport = True
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

            Move_Object_to_Hair_Curves_Collection(hair_cur)
            hair_cur_coll = hair_cur.users_collection[0]
            
            if bpy.app.version[0]==2 and bpy.app.version[1]>83 :
                hair_cur.data.bevel_mode = 'OBJECT'    
            elif bpy.app.version[0]>=3:
                hair_cur.data.bevel_mode = 'OBJECT'    
            
            hair_cur.data.bevel_depth = 0
            hair_cur.data.bevel_object = None
            hair_cur.data.taper_object = None


            cur_t = None
            for cur in cur_coll.objects:
                if cur.name[:5] == 'ANM_T':
                    cur_t = bpy.data.objects.new(cur.name, cur.data)
                    cur_t.data = cur_t.data.copy()
                    hair_cur_coll.objects.link(cur_t)
                    hair_cur.data.use_path = False
                    cur_t.parent = hair_cur
                    cur_t.hide_viewport = True
                    cur_t.hide_render = True
            for cur in bevel_coll.objects:
                if cur.type  == 'CURVE':
                    cur_b = bpy.data.objects.new(cur.name, cur.data)
                    cur_b.data = cur_b.data.copy()
                    
                    hair_cur_coll.objects.link(cur_b)
                    cur_t.data.use_path = False
                    cur_b.parent = cur_t

                    cur_b.hide_viewport = True
                    cur_b.hide_render = True


            geo_node.node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]

            bpy.context.view_layer.objects.active = hair_cur


            if not hair_cur.active_material:
                Import_ANM_Mat()
                mat = bpy.data.materials['ANM_Mat']
                hair_cur.data.materials.append(mat)

            geo_node.node_group.nodes['Set Material'].inputs[2].default_value =hair_cur.active_material
            geo_node["Output_12_attribute_name"] = "UVMap"


            bpy.ops.outliner.orphans_purge()

        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}


class Make_Curl_Hair_OP(Operator):
    """Make Curl Hair"""
    bl_label = "Make Curl Hair From Curve"
    bl_idname = "make_curl_hair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0] == obj and obj.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object
        animehair_props = context.scene.animehair_props
        
        if hair_cur.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

        braid_name = 'ANM_Curl_'
        if "ANM_Simple_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Simple_",braid_name)
        elif "ANM_Braid2_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid2_",braid_name)
        elif "ANM_Braid3_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid3_",braid_name)
        elif "ANM_Braid4_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid4_",braid_name)
        elif "ANM_Braid5_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid5_",braid_name)
        elif "ANM_Braid6_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_Braid6_",braid_name)
        elif "ANM_HairCurves_" in hair_cur.name:
            hair_cur.name = hair_cur.name.replace("ANM_HairCurves_",braid_name)
        elif not braid_name in hair_cur.name:
            hair_cur.name = braid_name + hair_cur.name


        for mod in hair_cur.modifiers:
            if mod.type == 'NODES':
                if 'ANM_' in mod.name:
                    hair_cur.modifiers.remove(mod)
        
        try:
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()

            elif bpy.data.collections.get('ANM_Curves_profile') and animehair_props.bool_source_file_updated:
                ANM_Curves_collection = bpy.data.collections.get('ANM_Curves_profile')
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()
                
                Import_curves_presets()

                animehair_props.bool_source_file_updated = False

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()


            script_file = os.path.realpath(__file__)
            directory = os.path.dirname(script_file)
            
            
            nod_gp_name = 'ANM_Curl_GN'
            filepath = os.path.join(directory , "ANM_Geo_Nodes.blend")  
            if not bpy.data.node_groups.get(nod_gp_name):
                with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                    data_to.node_groups = [
                        name for name in data_from.node_groups if name.startswith(nod_gp_name)]

            bpy.data.node_groups[nod_gp_name].use_fake_user = True

            geo_node = hair_cur.modifiers.new(nod_gp_name,'NODES')
            geo_node.name =  "ANM_Geo_Node"
            geo_node.node_group = bpy.data.node_groups[nod_gp_name]
            geo_node.node_group = geo_node.node_group.copy()
            
            for obj in hair_cur.children:
                if "ANM_Taper" in obj.name:
                    if len(obj.children):
                        for ob in obj.children:
                            bpy.data.objects.remove(ob, do_unlink = True)
                    bpy.data.objects.remove(obj, do_unlink = True)


            for cur in bpy.data.curves:
                if not cur.users:
                    bpy.data.curves.remove(cur)

            

            cur_coll = bpy.data.collections['ANM_Curves_profile']
            cur_coll.hide_viewport = True
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

            Move_Object_to_Hair_Curves_Collection(hair_cur)
            hair_cur_coll = hair_cur.users_collection[0]
            
            
            hair_cur.data.bevel_mode = 'OBJECT'    
            hair_cur.data.bevel_depth = 0
            hair_cur.data.bevel_object = None
            hair_cur.data.taper_object = None

            cur_t = None
            for cur in cur_coll.objects:
                if cur.name[:5] == 'ANM_T':
                    cur_t = bpy.data.objects.new(cur.name, cur.data)
                    cur_t.data = cur_t.data.copy()
                    hair_cur_coll.objects.link(cur_t)
                    hair_cur.data.use_path = False
                    cur_t.parent = hair_cur
                    cur_t.hide_viewport = True
                    cur_t.hide_render = True
            for cur in bevel_coll.objects:
                if cur.type  == 'CURVE':
                    cur_b = bpy.data.objects.new(cur.name, cur.data)
                    cur_b.data = cur_b.data.copy()
                    
                    hair_cur_coll.objects.link(cur_b)
                    cur_t.data.use_path = False
                    cur_b.parent = cur_t

                    cur_b.hide_viewport = True
                    cur_b.hide_render = True


            geo_node.node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]

            bpy.context.view_layer.objects.active = hair_cur

            if not hair_cur.active_material:
                Import_ANM_Mat()
                mat = bpy.data.materials['ANM_Mat']
                hair_cur.data.materials.append(mat)

            geo_node.node_group.nodes['Set Material'].inputs[2].default_value =hair_cur.active_material
            geo_node["Output_12_attribute_name"] = "UVMap"


            bpy.ops.outliner.orphans_purge()

        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}


class Anime_Hair_With_New_Hair_System_OP(Operator):
    """Add New Hair Curves to Mesh
(Make Sure mesh object has proper non-overlapping UV)"""

    bl_label = "Make Anime Hair From New Hair System"
    bl_idname = "anime_hair_with_new_hair_system.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        if len(objs)==1 :
            if objs[0].type == 'MESH' and context.active_object == objs[0] :
                c = True
        return c

    def execute(self, context):
        obj = context.active_object
        animehair_props =  context.scene.animehair_props
        if obj.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

        try:
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()

            elif bpy.data.collections.get('ANM_Curves_profile') and animehair_props.bool_source_file_updated:
                ANM_Curves_collection = bpy.data.collections.get('ANM_Curves_profile')
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()
                
                Import_curves_presets()

                animehair_props.bool_source_file_updated = False

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()

            cur_coll = bpy.data.collections['ANM_Curves_profile']
            cur_coll.hide_viewport = True
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']


            script_file = os.path.realpath(__file__)
            directory = os.path.dirname(script_file)
            nod_gp_name = 'ANM_HairCurves_GN'
            filepath = os.path.join(directory , "ANM_Geo_Nodes.blend")  
            if not bpy.data.node_groups.get(nod_gp_name):
                with bpy.data.libraries.load(filepath, link=False) as (data_from, data_to):
                    data_to.node_groups = [
                        name for name in data_from.node_groups if name.startswith(nod_gp_name)]
            bpy.data.node_groups[nod_gp_name].use_fake_user = True


            if obj.type == 'MESH':
                bpy.ops.object.curves_empty_hair_add()
                hair_curves = context.active_object
                
                hair_curves.parent= None
                hair_curves.name = "ANM_HairCurves_" + hair_curves.name

                Move_Object_to_Hair_Curves_Collection(hair_curves)

                hair_curves.select_set(True)
                bpy.context.view_layer.objects.active = hair_curves

                hair_cur_coll = hair_curves.users_collection[0]

                cur_t = None
                for cur in cur_coll.objects:
                    if cur.name[:5] == 'ANM_T':
                        cur_t = bpy.data.objects.new(cur.name, cur.data)
                        cur_t.data = cur_t.data.copy()
                        hair_cur_coll.objects.link(cur_t)
                        cur_t.parent = hair_curves
                        cur_t.hide_viewport = True
                        cur_t.hide_render = True
                for cur in bevel_coll.objects:
                    if cur.type  == 'CURVE':
                        cur_b = bpy.data.objects.new(cur.name, cur.data)
                        cur_b.data = cur_b.data.copy()
                        
                        hair_cur_coll.objects.link(cur_b)
                        cur_t.data.use_path = False
                        cur_b.parent = cur_t

                        cur_b.hide_viewport = True
                        cur_b.hide_render = True
                

                geo_node = hair_curves.modifiers.new(nod_gp_name,'NODES')
                geo_node.name = 'ANM_Geo_Node'
                geo_node.node_group = bpy.data.node_groups[nod_gp_name]
                geo_node.node_group = geo_node.node_group.copy()

                geo_node.node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[0]


                if not hair_curves.active_material:
                    Import_ANM_Mat()
                    mat = bpy.data.materials['ANM_Mat']
                    hair_curves.data.materials.append(mat)

                geo_node.node_group.nodes['Set Material'].inputs[2].default_value = hair_curves.active_material
                geo_node["Output_12_attribute_name"] = "UVMap"


                bpy.ops.object.mode_set(mode='SCULPT_CURVES')
                bpy.ops.wm.tool_set_by_id(name="builtin_brush.add")


        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')

        return {'FINISHED'}


class Reset_Hair_Curve_Tilt_OP(Operator):
    """Reset Hair Curve Tilt on Selected Points"""
    bl_label = "Reset Hair Curve Tilt"
    bl_idname = "reset_hair_curve_tilt.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0] == obj and obj.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object

        for spline in hair_cur.data.splines:
            if spline.type == 'BEZIER':
                for pt in spline.bezier_points:
                    if pt.select_control_point or pt.select_left_handle or pt.select_right_handle:
                        pt.tilt = 0
            elif spline.type == 'NURBS':
                for pt in spline.points:
                    if pt.select:
                        pt.tilt = 0
            elif spline.type == 'POLY':
                for pt in spline.points:
                    if pt.select:
                        pt.tilt = 0


        return {'FINISHED'}


class Reset_Hair_Curve_Radius_OP(Operator):
    """Reset Hair Curve Radius on Selected Points"""
    bl_label = "Reset Hair Curve Radius"
    bl_idname = "reset_hair_curve_radius.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0] == obj and obj.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object

        for spline in hair_cur.data.splines:
            if spline.type == 'BEZIER':
                for pt in spline.bezier_points:
                    if pt.select_control_point or pt.select_left_handle or pt.select_right_handle:
                        pt.radius = 1
            elif spline.type == 'NURBS':
                for pt in spline.points:
                    if pt.select:
                        pt.radius = 1
            elif spline.type == 'POLY':
                for pt in spline.points:
                    if pt.select:
                        pt.radius = 1

        return {'FINISHED'}



def not_enough_point_error(self, context):
    self.layout.label(text="need at lease 6 points for nurbs curves.")


class Simplify_Hair_OP(Operator):
    """simplify hair for edit"""
    bl_label = "Simplify Hair"
    bl_idname = "simplifyhair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        if context.active_object != None:
            if context.active_object.type == 'CURVE':
                c = True
        return c

    def execute(self, context):
        hair_cur = context.active_object
        if not hair_cur.select_get(): 
            hair_cur.select_set(True)

        if hair_cur.mode != "EDIT":
            bpy.ops.object.mode_set(mode='EDIT')

        condi_met = True
        splines = hair_cur.data.splines
        for spline in splines:
            if spline.type == 'NURBS':
                if len(spline.points)< 6:
                    condi_met = False
                    context.window_manager.popup_menu(not_enough_point_error, title="Error", icon='ERROR')
        if condi_met:
            bpy.ops.curve.select_all(action='SELECT')
            bpy.ops.curve.spline_type_set(type='BEZIER')
            bpy.ops.curve.handle_type_set(type='ALIGNED')

        return {'FINISHED'}

        

class AHM_Select_Main_Handle_OP(Operator):
    """Switch to main handle of selected bezier point"""
    bl_label = "Select Bezier Point Main Handle"
    bl_idname = "anm_select_main_handle.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True    
        return c


    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects

        for ob in objs:
            if ob.type !='CURVE':
                bpy.ops.object.mode_set(mode='OBJECT')
                ob.select_set(False)
                bpy.ops.object.mode_set(mode='EDIT')
                objs = context.selected_objects

        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'BEZIER' and not spline.hide :
                    for point in spline.bezier_points:
                        if point.select_control_point or point.select_left_handle or point.select_right_handle:
                            point.select_left_handle = False
                            point.select_right_handle = False
                            point.select_control_point = True
                    
            
        return {'FINISHED'}

class AHM_Select_Left_Handle_OP(Operator):
    """Switch to left handle of selected bezier point"""
    bl_label = "Select Bezier Point Left Handle"
    bl_idname = "anm_select_left_handle.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True    
        return c


    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects

        for ob in objs:
            if ob.type !='CURVE':
                bpy.ops.object.mode_set(mode='OBJECT')
                ob.select_set(False)
                bpy.ops.object.mode_set(mode='EDIT')
                objs = context.selected_objects

        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'BEZIER' and not spline.hide :
                    for point in spline.bezier_points:
                        if point.select_control_point or point.select_left_handle or point.select_right_handle:
                            point.select_left_handle = True
                            point.select_right_handle = False
                            point.select_control_point = False
                    
            
        return {'FINISHED'}

class AHM_Select_Right_Handle_OP(Operator):
    """Switch to right handle of selected bezier point"""
    bl_label = "Select Bezier Point Right Handle"
    bl_idname = "anm_select_right_handle.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True    
        return c


    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects
        for ob in objs:
            if ob.type !='CURVE':
                bpy.ops.object.mode_set(mode='OBJECT')
                ob.select_set(False)
                bpy.ops.object.mode_set(mode='EDIT')
                objs = context.selected_objects

        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'BEZIER' and not spline.hide :
                    for point in spline.bezier_points:
                        if point.select_control_point or point.select_left_handle or point.select_right_handle:
                            point.select_left_handle = False
                            point.select_right_handle = True
                            point.select_control_point = False
                    
        return {'FINISHED'}

class AHM_Isolate_Selected_Strand_Show_All_OP(Operator):
    """Isolate/show hair strand """
    bl_label = "Isolate Spline or Show All"
    bl_idname = "anm_isolate_selected_strand_show_all.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True
        return c

    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects

        for ob in objs:
            if ob.type !='CURVE':
                bpy.ops.object.mode_set(mode='OBJECT')
                ob.select_set(False)
                bpy.ops.object.mode_set(mode='EDIT')
                objs = context.selected_objects

        has_hidden_points = False

        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'BEZIER':
                     for point in spline.bezier_points:
                        if point.hide:
                            has_hidden_points=True
                            bpy.ops.curve.reveal()
                            bpy.ops.curve.select_all(action='SELECT')
                elif spline.type == 'NURBS' or spline.type == 'POLY':
                    for point in spline.points:
                        if point.hide:
                            has_hidden_points=True
                            bpy.ops.curve.reveal()
                            bpy.ops.curve.select_all(action='SELECT')

        if not has_hidden_points:
            bpy.ops.curve.select_linked()
            bpy.ops.curve.hide(unselected=True) 

        return {'FINISHED'}


class AHM_Hide_Selected_Strand_OP(Operator):
    """Hide Selected strand """
    bl_label = "Hide Selected Spline "
    bl_idname = "anm_hide_selected_strand.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True
        return c

    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)
        bpy.ops.curve.select_linked()
        bpy.ops.curve.hide(unselected=False)

        return {'FINISHED'}


class AHM_Selected_Next_Point_OP(Operator):
    """Select Next Point on Curve """
    bl_label = "Select Next Point On Curve "
    bl_idname = "anm_selected_next_point.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True    
        return c

    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects

        for ob in objs:
            if ob.type !='CURVE':
                bpy.ops.object.mode_set(mode='OBJECT')
                ob.select_set(False)
                bpy.ops.object.mode_set(mode='EDIT')
                objs = context.selected_objects

        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'NURBS' or spline.type == 'POLY':
                    if not spline.hide:
                        pts_in_spline = len(spline.points)
                        for i,point in enumerate(spline.points):
                            if not point.hide and point.select:
                                spline.points[i].select = False
                                next_pt_index = i+1
                                if next_pt_index > pts_in_spline -1:
                                    next_pt_index = pts_in_spline -1
                                spline.points[next_pt_index].select = True
                                break
                elif spline.type == 'BEZIER' and not spline.hide:
                    pts_in_spline = len(spline.bezier_points)
                    for i,point in enumerate(spline.bezier_points):
                        if not point.hide:
                            if point.select_control_point or point.select_left_handle or point.select_right_handle:
                                spline.bezier_points[i].select_control_point = False
                                spline.bezier_points[i].select_left_handle = False
                                spline.bezier_points[i].select_right_handle = False
                                next_pt_index = i+1
                                if next_pt_index > pts_in_spline -1:
                                    next_pt_index = pts_in_spline -1
                                spline.bezier_points[next_pt_index].select_control_point = True
                                break

        return {'FINISHED'}


class AHM_Selected_Prev_Point_OP(Operator):
    """Select Previous Point on Curve """
    bl_label = "Select Previous Point On Curve"
    bl_idname = "anm_selected_prev_point.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True    
        return c

    def execute(self, context):

        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects

        for ob in objs:
            if ob.type !='CURVE':
                bpy.ops.object.mode_set(mode='OBJECT')
                ob.select_set(False)
                bpy.ops.object.mode_set(mode='EDIT')
                objs = context.selected_objects

        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'NURBS' or spline.type == 'POLY':
                    if  not spline.hide:
                        pts_in_spline = len(spline.points)
                        for i,point in enumerate(spline.points):
                            if not point.hide and point.select:
                                spline.points[i].select = False
                                prev_pt_index = i-1
                                if prev_pt_index < 0:
                                    prev_pt_index = 0
                                spline.points[prev_pt_index].select = True
                                break
                elif spline.type == 'BEZIER'and not spline.hide:
                    pts_in_spline = len(spline.bezier_points)
                    for i,point in enumerate(spline.bezier_points):
                        if not point.hide:
                            if point.select_control_point or point.select_left_handle or point.select_right_handle:
                                spline.bezier_points[i].select_control_point = False
                                spline.bezier_points[i].select_left_handle = False
                                spline.bezier_points[i].select_right_handle = False
                                prev_pt_index = i-1
                                if prev_pt_index < 0:
                                    prev_pt_index = 0
                                spline.bezier_points[prev_pt_index].select_control_point = True
                                break
        return {'FINISHED'}




def not_saved_warning_error(self, context):
    self.layout.label(text="Save Blend File Before Proceed.")

class Add_Curve_To_Source_File_OP (Operator):
    """Append Currently Selected Curve To Source Bevel Profile
Make Sure It Faces Up Z Direction and not parent or child of any object"""
    bl_label = "Append Active Curve As Bevel Profle To Source File?"
    bl_idname = "add_curve_to_source_file.operator"

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs)==1:
            if  objs[0] == obj and  obj.type == 'CURVE':
                c = True
        return c

    def invoke(self, context, event):

        return context.window_manager.invoke_props_dialog(self)


    def execute(self, context):
        obj = bpy.context.active_object
        pref_path = context.preferences.addons[__package__].preferences.filepath
        if obj.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
        try:
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)
        except:
            self.report({'ERROR'}, "曲线需为3D不能为2D!")
            return {"CANCELLED"}
        if not bpy.data.is_saved:
            context.window_manager.popup_menu(not_saved_warning_error, title="Error", icon='ERROR')
        else:
            try:
                bpy.ops.wm.save_mainfile()
                obj.name = 'NEW_ANM_Bevel_profile'
                
                if len(obj.children):
                    for ob in obj.children:
                        bpy.data.objects.remove(ob, do_unlink=True)
                
                bpy.ops.view3d.copybuffer()
                source_directory = ''
                if pref_path == '':
                    script_file = os.path.realpath(__file__)
                    source_directory = os.path.dirname(script_file)
                else:
                    source_directory = pref_path

                source_file_path = os.path.join(source_directory , "Curves_Presets.blend")  
                bpy.ops.wm.open_mainfile(filepath= source_file_path)

                bpy.ops.view3d.pastebuffer()
                new_cur = bpy.data.objects['NEW_ANM_Bevel_profile']
                new_cur.name = new_cur.name.replace("NEW_", "")
                
                cur_collection = bpy.data.collections['ANM_Bevel_Profiles']
                
                new_cur.users_collection[0].objects.unlink(new_cur)

                cur_collection.objects.link(new_cur)

                bpy.ops.wm.save_mainfile()

                if bpy.app.version[0] ==2:
                    fp = bpy.utils.user_resource('CONFIG', "recent-files.txt")
                elif bpy.app.version[0] >=3:
                    fp = bpy.utils.user_resource('CONFIG',  path="recent-files.txt")

                try:
                    with open(fp) as file:
                        recent_files = file.read().splitlines()
                except (IOError, OSError, FileNotFoundError):
                    recent_files = []
                    
                bpy.ops.wm.open_mainfile(filepath= recent_files[0])

                context.scene.animehair_props.bool_source_file_updated = True
            except:
                bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')

        return {'FINISHED'}



class Finalize_ANM_Hair_OP(Operator):
    """Turn selected Simple Curves into meshes"""
    bl_label = "Finalize ANM Hair"
    bl_idname = "anm_finalize_hair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs):
            if obj.type == 'CURVE':
                c = True    
        return c

    def execute(self, context):
        objs = context.selected_objects
        animehair_props =  context.scene.animehair_props

        if objs[0].mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        is_cap = False

        for ob in objs:
            ob.select_set(False)
            if ob.type == 'CURVE' and 'ANM_Simple_' in ob.name:
                ob.select_set(True)
                
                if animehair_props.fill_cap:
                    ob.data.use_fill_caps = False
                    is_cap = True
        objs = context.selected_objects

        if len(objs):
            bpy.ops.object.duplicate()

            for ob in objs:
                ob.hide_set(True)
                ob.hide_viewport = True
                ob.hide_render = True

            
            bpy.ops.object.convert(target='MESH')
            objs = context.selected_objects
            for obj in objs:
                obj.name = obj.name.replace('ANM_','')
                Move_Object_to_Hair_Curves_Collection(obj)
            
            bpy.context.space_data.overlay.show_wireframes = False

            if animehair_props.auto_merge_end:
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.select_mode(type="VERT")
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.remove_doubles(threshold=0.00001)
                bpy.ops.object.mode_set(mode='OBJECT')

            if animehair_props.root_vg:
                bpy.ops.mesh.separate(type='LOOSE')
                objs = bpy.context.selected_objects
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.select_mode(type="VERT")
                bpy.ops.mesh.select_all(action='DESELECT')
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.select_all(action='DESELECT')
                for obj in objs:
                    obj.select_set(True)
                    bpy.context.view_layer.objects.active = obj
                    if animehair_props.taper_objects_root > 0:
                        obj.data.vertices[0].select = True
                        obj.data.vertices[1].select = True
                        bpy.ops.object.mode_set(mode='EDIT')
                        bpy.ops.mesh.loop_multi_select(ring=False)
                    elif animehair_props.taper_objects_root == 0:
                        obj.data.vertices[0].select = True
                        bpy.ops.object.mode_set(mode='EDIT')
                    pin_vg = obj.vertex_groups.new(name='Root_Pin')
                    bpy.ops.object.vertex_group_assign()

                    if is_cap:
                        bpy.ops.mesh.select_all(action='SELECT')
                        bpy.ops.mesh.region_to_loop()
                        bm = bmesh.from_edit_mesh(obj.data)
                        verts = [ v.index for v in bm.verts if v.select ]
                        if len(verts):
                            bpy.ops.mesh.fill()
                            bpy.ops.mesh.normals_make_consistent(inside=False)

                    obj.select_set(False)

                bpy.ops.object.mode_set(mode='OBJECT')
            
            elif animehair_props.root_vg == False and is_cap == True:
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.select_mode(type="VERT")
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.region_to_loop()
                bm = bmesh.from_edit_mesh(obj.data)
                verts = [ v.index for v in bm.verts if v.select ]
                if len(verts):
                    bpy.ops.mesh.fill()
                    bpy.ops.mesh.normals_make_consistent(inside=False)

        
        return {'FINISHED'}


class Finalize_Curl_Braids_Hair_OP(Operator):
    """Turn selected Curl and Braids Hair into meshes"""
    bl_label = "Finalize Curl and Braids Hair"
    bl_idname = "anm_finalize_curl_braids_hair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs):
            if obj.type == 'CURVE' :
                c = True    
        return c

    def execute(self, context):
        obj = context.active_object
        objs = context.selected_objects
        animehair_props =  context.scene.animehair_props

        if not obj.select_get():
            obj.select_set(True)

        if obj.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        for ob in objs:
            ob.select_set(False)
            if ob.type == 'CURVE':
                if 'ANM_Braid' in ob.name or 'ANM_Curl' in ob.name:
                    ob.select_set(True)
                    
                    if animehair_props.fill_cap and ob.modifiers.get('ANM_Geo_Node'):
                        ob.modifiers['ANM_Geo_Node'].node_group.nodes["Curve to Mesh"].inputs[2].default_value = True
                        
                    elif animehair_props.fill_cap == False and ob.modifiers.get('ANM_Geo_Node'):
                        ob.modifiers['ANM_Geo_Node'].node_group.nodes["Curve to Mesh"].inputs[2].default_value = False

        objs = context.selected_objects
        if len(objs):
            bpy.ops.object.duplicate()

            for ob in objs:
                ob.hide_set(True)
                ob.hide_viewport = True
                ob.hide_render = True

            
            bpy.ops.object.convert(target='MESH')
            objs = context.selected_objects
            for obj in objs:
                obj.name = obj.name.replace('ANM_','')
                Move_Object_to_Hair_Curves_Collection(obj)



            bpy.context.space_data.overlay.show_wireframes = False

            if animehair_props.auto_merge_end:
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.select_mode(type="VERT")
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.remove_doubles(threshold=0.00001)
                bpy.ops.object.mode_set(mode='OBJECT')

   
        return {'FINISHED'}



class Convert_Hair_Curves_to_Simple_Curve_OP(Operator):
    """Convert Hair Curves to Curve for additional editing"""
    bl_label = "Convert Hair Curves to Simple Curve"
    bl_idname = "anm_curves_to_simple_curve.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs)==1:
            if objs[0]== obj and obj.type == 'CURVES':
                c = True    
        return c

    def execute(self, context):
        obj = context.active_object
        animehair_props =  context.scene.animehair_props
        mat = obj.active_material
        if obj.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        surface_mesh = obj.data.surface

        if surface_mesh and obj.modifiers.get('ANM_Geo_Node'):
            geo_node = obj.modifiers['ANM_Geo_Node']
            res_length = geo_node["Input_8"]
            res_profile = geo_node.node_group.nodes["Object Info"].inputs[0].default_value.data.resolution_u 
            strand_rad = geo_node.node_group.nodes["Value"].outputs[0].default_value
            strand_tilt = geo_node.node_group.nodes['Transform'].inputs[2].default_value[2]
            sub_d = obj.modifiers["ANM_Geo_Node"]["Input_10"]

            profile_mesh = geo_node.node_group.nodes["Object Info"].inputs[0].default_value
            res_profile = profile_mesh.data.resolution_u 

            profile_mesh_indx = 0
            org_cur_t = None
            for ob in obj.children:
                if 'ANM_Taper' in ob.name:
                    org_cur_t = ob
                    break
            if org_cur_t and len(org_cur_t.children):
                for i, ob in enumerate(org_cur_t.children):
                    if ob.name == profile_mesh.name:
                        profile_mesh_indx = i
                        break

            
            bpy.ops.curves.convert_to_particle_system()
            
            obj.select_set(False)
            obj.hide_set(True)
            obj.hide_viewport = True
            obj.hide_render = True
            surface_mesh.select_set(True)
            bpy.context.view_layer.objects.active = surface_mesh
            
            surface_mesh.particle_systems.active_index = len(surface_mesh.particle_systems)-1

            bpy.ops.object.modifier_convert(modifier=obj.name)

            new_hair = context.active_object

            new_hair.select_set(False)
            surface_mesh.select_set(True)
            bpy.context.view_layer.objects.active = surface_mesh
            bpy.ops.object.particle_system_remove()
            surface_mesh.select_set(False)

            new_hair.select_set(True)
            bpy.context.view_layer.objects.active = new_hair
            bpy.ops.object.convert(target='CURVE')

            Move_Object_to_Hair_Curves_Collection(new_hair)
            bpy.context.view_layer.objects.active = new_hair
            

            try:
                if not bpy.data.collections.get('ANM_Curves_profile'):
                    Import_curves_presets()

                elif bpy.data.collections.get('ANM_Curves_profile') and animehair_props.bool_source_file_updated:
                    ANM_Curves_collection = bpy.data.collections.get('ANM_Curves_profile')
                    for obj in ANM_Curves_collection.objects:
                        bpy.data.objects.remove(obj, do_unlink=True)
                    bpy.data.collections.remove(ANM_Curves_collection)
                    bpy.ops.outliner.orphans_purge()
                    
                    Import_curves_presets()

                    animehair_props.bool_source_file_updated = False

                elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                    ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                    for obj in ANM_Curves_collection.objects:
                        bpy.data.objects.remove(obj, do_unlink=True)
                    bpy.data.collections.remove(ANM_Curves_collection)
                    bpy.ops.outliner.orphans_purge()

                    Import_curves_presets()
            except:
                bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')

                
            new_hair.name = 'ANM_HairCurves_Curves'

            geo_node = new_hair.modifiers.new('ANM_Geo_Node','NODES')
            geo_node.node_group = obj.modifiers['ANM_Geo_Node'].node_group
            geo_node.node_group = geo_node.node_group.copy()

            hair_cur_coll = new_hair.users_collection[0]
            cur_coll = bpy.data.collections['ANM_Curves_profile']
            cur_coll.hide_viewport = True
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

            
            cur_t = None
            for cur in cur_coll.objects:
                if cur.name[:5] == 'ANM_T':
                    cur_t = bpy.data.objects.new(cur.name, cur.data)
                    cur_t.data = cur_t.data.copy()
                    hair_cur_coll.objects.link(cur_t)
                    new_hair.data.use_path = False
                    cur_t.parent = new_hair 

                    cur_t.hide_viewport = True
                    cur_t.hide_render = True
            for cur in bevel_coll.objects:
                if cur.type  == 'CURVE':
                    cur_b = bpy.data.objects.new(cur.name, cur.data)
                    cur_b.data = cur_b.data.copy()
                    
                    hair_cur_coll.objects.link(cur_b)
                    cur_t.data.use_path = False
                    cur_b.parent = cur_t

                    cur_b.hide_viewport = True
                    cur_b.hide_render = True


            geo_node.node_group.nodes["Object Info"].inputs[0].default_value = cur_t.children[profile_mesh_indx]

            geo_node.node_group.nodes["Switch"].inputs[0].default_value = True

            new_hair.modifiers['ANM_Geo_Node']["Input_8"] = res_length
            new_hair.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value.data.resolution_u = res_profile 
            new_hair.modifiers['ANM_Geo_Node'].node_group.nodes["Value"].outputs[0].default_value = strand_rad 
            new_hair.modifiers['ANM_Geo_Node'].node_group.nodes['Transform'].inputs[2].default_value[2] = strand_tilt
            new_hair.modifiers["ANM_Geo_Node"]["Input_10"] = sub_d 
            new_hair.update_tag()


            if not new_hair.active_material:
                Import_ANM_Mat()
                new_hair.data.materials.append(mat)

            geo_node.node_group.nodes['Set Material'].inputs[2].default_value = mat
            geo_node["Output_12_attribute_name"] = "UVMap"

            
            bpy.ops.outliner.orphans_purge()
   
        return {'FINISHED'}


class Convert_Hair_Curves_to_Mesh_OP(Operator):
    """Convert Hair System to Mesh"""
    bl_label = "Convert Hair Curves to Mesh"
    bl_idname = "anm_curves_to_mesh.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs)==1:
            if objs[0] == obj and obj.type == 'CURVE':
                c = True    
        return c

    def execute(self, context):
        obj = context.active_object
        animehair_props =  context.scene.animehair_props

        if obj.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        if obj.type == 'CURVE':
            if 'ANM_HairCurves' in obj.name:

                bpy.ops.object.duplicate()
                obj_mesh = context.active_object

                obj.hide_set(True)
                obj.hide_viewport = True
                obj.hide_render = True

                obj_mesh.name = obj_mesh.name.replace('ANM_','')

                if animehair_props.fill_cap and obj_mesh.modifiers.get('ANM_Geo_Node'):
                    obj_mesh.modifiers['ANM_Geo_Node'].node_group.nodes["Curve to Mesh"].inputs[2].default_value = True
                    
                elif animehair_props.fill_cap == False and obj_mesh.modifiers.get('ANM_Geo_Node'):
                    obj_mesh.modifiers['ANM_Geo_Node'].node_group.nodes["Curve to Mesh"].inputs[2].default_value = False

                bpy.ops.object.convert(target='MESH')
                
                Move_Object_to_Hair_Curves_Collection(obj_mesh)

                bpy.context.space_data.overlay.show_wireframes = False

                if animehair_props.auto_merge_end:
                    bpy.ops.object.mode_set(mode='EDIT')
                    bpy.ops.mesh.select_mode(type="VERT")
                    bpy.ops.mesh.select_all(action='SELECT')
                    bpy.ops.mesh.remove_doubles(threshold=0.00001)
                    bpy.ops.object.mode_set(mode='OBJECT')

   
        return {'FINISHED'}






class Edit_ANM_Hair_Bevel_Object_OP(Operator):
    """Isolate and Edit Hair Bevel Object Shape"""
    bl_label = "Edit_ANM_Hair_Bevel_Object"
    bl_idname = "anm_edit_hair_bevel_object.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs)==1:
            if obj == objs[0] and obj.type =='CURVE':
                c = True
        return c

    def execute(self, context):
        obj = context.active_object
        if obj.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        if obj.type == 'CURVE':
            bevel_obj = obj.data.bevel_object
            bevel_obj.hide_viewport = False
            bevel_obj.select_set(True)
            bpy.context.view_layer.objects.active = bevel_obj
            obj.select_set(False)
            bpy.ops.view3d.view_selected(use_all_regions=False)
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.curve.select_all(action='SELECT')
            bpy.context.space_data.shading.show_backface_culling = False

        return {'FINISHED'}


class Edit_ANM_Hair_Geo_Node_Profile_Object_OP(Operator):
    """Isolate and Edit Hair Bevel Object Shape"""
    bl_label = "Edit_ANM_Hair_Bevel_Object"
    bl_idname = "anm_edit_hair_geo_node_profile_object.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs)==1 and obj == objs[0]:
            if obj.type =='CURVE' or obj.type =='CURVES':
                c = True
        
        return c

    def execute(self, context):
        obj = context.active_object
        if obj.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')
        
        if obj.modifiers.get('ANM_Geo_Node'):
            profile_cur = obj.modifiers['ANM_Geo_Node'].node_group.nodes["Object Info"].inputs[0].default_value

            if obj.type == 'CURVES':
                surface_mesh = obj.data.surface
                surface_mesh.hide_viewport = True

            profile_cur.hide_viewport = False
            profile_cur.select_set(True)
            bpy.context.view_layer.objects.active = profile_cur
            obj.select_set(False)
            bpy.ops.view3d.view_selected(use_all_regions=False)
            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.curve.select_all(action='SELECT')
            bpy.context.space_data.shading.show_backface_culling = False

        return {'FINISHED'}

class Hide_ANM_Hair_Bevel_Objects_OP(Operator):
    """Exit bevel profile edit, restore to anime hair view"""
    bl_label = "Edit_ANM_Hair_Bevel_Object"
    bl_idname = "anm_hide_hair_bevel_objects.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        obj = context.active_object
        if len(objs)==1:
            if obj == objs[0] and obj.type =='CURVE':
                c = True
        return c

    def execute(self, context):
        obj = context.active_object

        if obj.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        obj.hide_viewport = True

        hair_cur = obj.parent
        if hair_cur:
            if "ANM_Taper" in hair_cur.name:
                if hair_cur.parent.type == 'CURVES':
                    surface_mesh = hair_cur.parent.data.surface
                    surface_mesh.hide_viewport = False
                hair_cur.parent.select_set(True)
                bpy.context.view_layer.objects.active = hair_cur.parent
                bpy.ops.view3d.view_selected(use_all_regions=False)
            elif 'ANM_Curl' in hair_cur.name or 'ANM_Braid' in hair_cur.name or 'ANM_HairCurves'in hair_cur.name:
                if hair_cur.type == 'CURVES':
                    surface_mesh = hair_cur.data.surface
                    surface_mesh.hide_viewport = False
                hair_cur.select_set(True)
                bpy.context.view_layer.objects.active = hair_cur
                bpy.ops.view3d.view_selected(use_all_regions=False)

        return {'FINISHED'}


class Create_Bones_From_Curves_OP(Operator):
    """Create Bones From Curve Points, select a curve to start"""
    bl_label = "Create Bones From Curve Points"
    bl_idname = "anm_create_bones_from_curves.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0].type == 'CURVE' and objs[0] == context.active_object:
                c = True
        return c

    def execute(self, context):
        cur = context.active_object
        cur_data = cur.data
        bpy.ops.object.mode_set(mode='OBJECT')
        arm_data = bpy.data.armatures.new("Hair_Armature")
        arm_obj = bpy.data.objects.new("Hair_Armature", arm_data)
        bpy.context.scene.collection.objects.link(arm_obj)
        arm_obj.show_in_front = True

        bpy.context.view_layer.objects.active = arm_obj
        arm_obj.matrix_world = cur.matrix_world
        bpy.ops.object.mode_set(mode='EDIT')

        for spline in cur_data.splines:
            if spline.type == 'BEZIER':
                parent = None
                pt_index = 0
                for i, bezier_point in enumerate(spline.bezier_points):
                    bone = arm_data.edit_bones.new("Hair_Bone")
                    bone.head = spline.bezier_points[i].co
                    bone.tail = spline.bezier_points[i+1].co
                    bone.parent = parent
                    bone.use_connect = True
                    parent = bone
                    pt_index = pt_index + 1
                    if pt_index > len(spline.bezier_points)-2:
                        break

            elif spline.type == 'NURBS' or spline.type == 'POLY':
                parent = None
                pt_index = 0
                for i, point in enumerate(spline.points):
                    bone = arm_data.edit_bones.new("Hair_Bone")
                    v=[0,0,0]
                    v[0] = spline.points[i].co[0]
                    v[1] = spline.points[i].co[1]
                    v[2] = spline.points[i].co[2]
                    bone.head = v
                    v[0] = spline.points[i+1].co[0]
                    v[1] = spline.points[i+1].co[1]
                    v[2] = spline.points[i+1].co[2]
                    bone.tail = v
                    bone.parent = parent
                    bone.use_connect = True
                    parent = bone
                    pt_index = pt_index + 1
                    if pt_index > len(spline.points)-2:
                        break
        
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.context.view_layer.objects.active = cur

        return {'FINISHED'}


class Add_ANM_Hair_On_Target_OP(Operator):
    """Add ANM Hair On Mesh Surface, select a mesh to start"""
    bl_label = "Add Hair On Target Surface"
    bl_idname = "anm_add_hair_on_target.operator"
    bl_options = {"REGISTER", "UNDO", "BLOCKING"}

    hair_list =[]
    source_hair = None
    
    target_surface = None

    @ classmethod
    def poll(cls, context):
        c = False
        objs = context.selected_objects
        if len(objs)==1:
            if objs[0].type == 'MESH' and context.active_object == objs[0] :
                c = True
        return c

    def prep(self, context):
        
        animehair_props =  context.scene.animehair_props
        self.hair_list = []
        self.target_surface = context.active_object

        try:
            if self.target_surface.mode != 'OBJECT':
                bpy.ops.object.mode_set(mode='OBJECT')
            self.target_surface.select_set(False)

            bpy.ops.wm.tool_set_by_id(name="builtin.select_box")
            
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()
                bpy.data.collections['ANM_Curves_profile'].hide_viewport = True

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()
                bpy.data.collections['ANM_Curves_profile'].hide_viewport = True


            if animehair_props.use_bezier:
                self.source_hair = bpy.data.objects.get('ANM_bezier')
            elif animehair_props.use_nurbs:
                self.source_hair = bpy.data.objects.get('ANM_nurbs')
        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')


    def finish_up(self, context):
        
        if self.temp:
            bpy.data.objects.remove(self.temp, do_unlink=True)

        if len(self.hair_list):
            for obj in self.hair_list:
                obj.select_set(True)
            
            bpy.context.view_layer.objects.active = self.hair_list[0]
            bpy.ops.object.join()
            bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

            hair_cur = context.active_object

            cur_coll = bpy.data.collections['ANM_Curves_profile']
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

            Move_Object_to_Hair_Curves_Collection(hair_cur)
            hair_cur_coll = hair_cur.users_collection[0]
            bpy.ops.object.shade_smooth()
            
            if bpy.app.version[0]==2 and bpy.app.version[1]>83 :
                hair_cur.data.bevel_mode = 'OBJECT'    
            elif bpy.app.version[0]>=3:
                hair_cur.data.bevel_mode = 'OBJECT'    

            cur_t = None
            for cur in cur_coll.objects:
                if cur.name[:5] == 'ANM_T':
                    cur_t = bpy.data.objects.new(cur.name, cur.data)
                    cur_t.data = cur_t.data.copy()
                    hair_cur_coll.objects.link(cur_t)
                    hair_cur.data.use_path = False
                    cur_t.parent = hair_cur
                    cur_t.hide_viewport = True
                    cur_t.hide_render = True
            for cur in bevel_coll.objects:
                if cur.type  == 'CURVE':
                    cur_b = bpy.data.objects.new(cur.name, cur.data)
                    cur_b.data = cur_b.data.copy()
                    
                    hair_cur_coll.objects.link(cur_b)
                    cur_t.data.use_path = False
                    cur_b.parent = cur_t
                    cur_b.hide_viewport = True
                    cur_b.hide_render = True

            hair_cur.data.taper_object = cur_t
            hair_cur.data.bevel_object = cur_t.children[0]

            bpy.context.view_layer.objects.active = hair_cur

            if not hair_cur.active_material:
                Import_ANM_Mat()
                mat = bpy.data.materials['ANM_Mat']
                hair_cur.data.materials.append(mat)


            Apply_existing_taper_value(hair_cur)

    def canceled_op(self, context):
        if self.temp:
            bpy.data.objects.remove(self.temp, do_unlink=True)

        if len(self.hair_list):
            for obj in self.hair_list:
                bpy.data.objects.remove(obj, do_unlink = True)

    def remove_recent_hair_op(self, context):
        if len(self.hair_list)>0:
            
            n = len(self.hair_list)-1
            ob = self.hair_list[n]
            self.hair_list.remove(ob)
            bpy.data.objects.remove(ob, do_unlink = True)

    def add_hair(self, context):
        if self.source_hair:
            hair_new = bpy.data.objects.new('ANM_Simple_Hairs', self.source_hair.data) 
            hair_new.data = hair_new.data.copy()                           
            self.target_surface.users_collection[0].objects.link(hair_new)             
            
            hair_new.matrix_world = self.temp.matrix_world
            
            self.hair_list.append(hair_new)
            

    def invoke(self, context, event):

        try:
            self.prep(context)

            self.mouse_pos = [0,0]
            self.loc = [0,0,0]
            self.view_point = None
            self.view_vector = None
            self.world_loc = None
            self.loc_on_surface = None
            self.hit_on_target = False
            
            self.temp = bpy.data.objects.new('hair_temp', self.source_hair.data)
            self.temp.data = self.temp.data.copy()
            self.target_surface.users_collection[0].objects.link(self.temp)
            bpy.context.view_layer.objects.active = self.temp
            self.temp.select_set(True)

            context.window_manager.modal_handler_add(self)
            context.workspace.status_text_set("LMB: Place Hair          " "Spacebar/Enter: Confirm         " 
            "Esc/RMB: Cancel       ""W and S: Scale Hair        " "Z: Remove Last Added Hair        ")

        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')
            return {"CANCELLED"}
        return {"RUNNING_MODAL"}


    def modal(self, context, event):

        if event.type in ['MIDDLEMOUSE' ,'WHEELUPMOUSE', 'WHEELDOWNMOUSE'] or event.alt or event.shift:
            return {'PASS_THROUGH'}

        elif event.type == 'W' and event.value == 'PRESS': 
            if self.temp:
                self.temp.scale[2] += 0.05
        elif event.type == 'S' and event.value == 'PRESS': 
            if self.temp:
                if self.temp.scale[2] > 0.1:
                    self.temp.scale[2] -= 0.05
        
        elif event.type == 'MOUSEMOVE':    
            self.mouse_pos = event.mouse_region_x, event.mouse_region_y
            region = bpy.context.region
            region3D = bpy.context.space_data.region_3d
            
            self.view_vector = view3d_utils.region_2d_to_vector_3d(region, region3D, self.mouse_pos)
            
            self.view_point = view3d_utils.region_2d_to_origin_3d(region, region3D, self.mouse_pos)
            
            self.world_loc = view3d_utils.region_2d_to_location_3d(region, region3D, self.mouse_pos, self.view_vector)
            self.loc_on_surface = None
            z = Vector( (0,0,1) )
            
            self.normal = z
            if self.target_surface:
                world_mat_inv = self.target_surface.matrix_world.inverted()
                rc_origin = world_mat_inv @ self.view_point
                rc_destination = world_mat_inv @ self.world_loc
                rc_direction = (rc_destination - rc_origin).normalized()
                hit, loc, norm, index = self.target_surface.ray_cast( origin = rc_origin, direction = rc_direction )
                self.loc_on_surface = loc
                if hit:
                    self.world_loc = self.target_surface.matrix_world @ loc
                    norm.rotate( self.target_surface.matrix_world.to_euler('XYZ') )    
                    self.normal = norm.normalized()
                    self.hit_on_target = True
                else:
                    self.hit_on_target = False
            if self.temp:
                self.temp.location = self.world_loc
                self.temp.rotation_euler = z.rotation_difference( self.normal ).to_euler()      
               
        elif event.type == 'LEFTMOUSE' and event.value == 'RELEASE': 
            if self.hit_on_target:
                self.add_hair(context)
        
        elif event.type == 'Z' and event.value == 'PRESS': 
            self.remove_recent_hair_op(context)

        elif event.type in {'RET', 'NUMPAD_ENTER', 'SPACE'}:  
            self.finish_up(context)
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            self.canceled_op(context)
            return {'CANCELLED'}
            
        return {"RUNNING_MODAL"}




class Draw_Bezier_Hairs_OP (Operator):
    """Draw Bezier Hairs"""
    bl_label = "Draw Bezier Hairs"
    bl_idname = "anm_draw_bezier_hairs.operator"

    def execute(self, context):
        try:
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()
                bpy.data.collections['ANM_Curves_profile'].hide_viewport = True

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()
                bpy.data.collections['ANM_Curves_profile'].hide_viewport = True

            bpy.ops.curve.primitive_bezier_curve_add()

            hair_cur = context.active_object
            if not "ANM_Simple_" in hair_cur.name:
                hair_cur.name = "ANM_Simple_" + hair_cur.name

            cur_coll = bpy.data.collections['ANM_Curves_profile']
            bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

            Move_Object_to_Hair_Curves_Collection(hair_cur)
            hair_cur_coll = hair_cur.users_collection[0]
            bpy.ops.object.shade_smooth()
            
            if bpy.app.version[0]==2 and bpy.app.version[1]>83 :
                hair_cur.data.bevel_mode = 'OBJECT'    
            elif bpy.app.version[0]>=3:
                hair_cur.data.bevel_mode = 'OBJECT'  
            
            cur_t = None
            for cur in cur_coll.objects:
                if cur.name[:5] == 'ANM_T':
                    cur_t = bpy.data.objects.new(cur.name, cur.data)
                    cur_t.data = cur_t.data.copy()
                    hair_cur_coll.objects.link(cur_t)
                    hair_cur.data.use_path = False
                    cur_t.parent = hair_cur 

                    cur_t.hide_viewport = True
                    cur_t.hide_render = True
            for cur in bevel_coll.objects:
                if cur.type  == 'CURVE':
                    cur_b = bpy.data.objects.new(cur.name, cur.data)
                    cur_b.data = cur_b.data.copy()
                    
                    hair_cur_coll.objects.link(cur_b)
                    cur_t.data.use_path = False
                    cur_b.parent = cur_t

                    cur_b.hide_viewport = True
                    cur_b.hide_render = True

            hair_cur.data.taper_object = cur_t
            hair_cur.data.bevel_object = cur_t.children[0]

            bpy.context.view_layer.objects.active = hair_cur

            if not hair_cur.active_material:
                Import_ANM_Mat()
                mat = bpy.data.materials['ANM_Mat']
                hair_cur.data.materials.append(mat)
            

            Apply_existing_taper_value(hair_cur)

            bpy.ops.object.mode_set(mode='EDIT')
            bpy.ops.curve.select_all(action='SELECT')
            bpy.ops.curve.delete(type='VERT')
            bpy.ops.wm.tool_set_by_id(name="builtin.draw")
            bpy.context.scene.tool_settings.curve_paint_settings.depth_mode = 'SURFACE'
            bpy.context.scene.tool_settings.curve_paint_settings.use_stroke_endpoints = True
            bpy.context.scene.tool_settings.curve_paint_settings.surface_plane = 'VIEW'
        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')

        return {'FINISHED'}




class Move_Curve_Point_OP(Operator):
    """Move Curve Point"""
    bl_label = "Move Curve Point"
    bl_idname = "anm_move_curve_point.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVE' and obj.mode == "EDIT":
                c = True    
        return c

    def execute(self, context):
        obj = context.active_object
        if not obj.select_get(): 
            obj.select_set(True)

        objs = context.selected_objects

        point_selected = False
        for ob in objs:
            for spline in ob.data.splines:
                if spline.type == 'NURBS' or spline.type == 'POLY':
                    for point in spline.points:
                        if point.select:
                            point_selected = True
                            break
                elif spline.type == 'BEZIER':
                    for point in spline.bezier_points:
                        if point.select_control_point:
                            point_selected = True
                            break
                    
        if point_selected:
            bpy.ops.transform.translate('INVOKE_DEFAULT')
            bpy.ops.curve.select_all(action='DESELECT')
        else:
            bpy.ops.view3d.select('INVOKE_DEFAULT')
            bpy.ops.transform.translate('INVOKE_DEFAULT')
            bpy.ops.curve.select_all(action='DESELECT')

        return {'FINISHED'}


def empty_path(self, context):
    self.layout.label(text="Empty file path. Locate the correct folder that contains Curves_Presets.blend.")

class Refresh_Bevel_Profiles_OP(Operator):
    """Refresh Bevel Profiles From File Path"""
    bl_label = "Refresh Bevel Profiles"
    bl_idname = "anm_refresh_bevel_profiles.operator"
    bl_options = {'REGISTER', 'UNDO'}

    def execute(self, context):
        path = context.preferences.addons[__package__].preferences.filepath
        collections = bpy.data.collections
        if path != '':
            try:
                if collections.get('ANM_Curves_profile'):
                    collections.remove(collections['ANM_Curves_profile'])
                    bpy.ops.outliner.orphans_purge()
                    Import_curves_presets()
                    collections['ANM_Curves_profile'].hide_viewport = True
                    context.scene.animehair_props.bool_source_file_updated = True
            except:
                bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')
        else:
            bpy.context.window_manager.popup_menu(empty_path)
        return {'FINISHED'}




class Make_ANM_Mat_Unique_OP(Operator):
    """Make selected hair material unique"""
    bl_label = "Make Mat Unique"
    bl_idname = "anm_make_mat_unique.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):   
        return context.active_object and len(context.selected_objects) == 1


    def execute(self, context):
        obj = context.active_object
        if obj.active_material and 'ANM_Mat' in obj.active_material.name:
            new_mat = obj.active_material.copy()
            obj.active_material = new_mat
            if obj.modifiers.get('ANM_Geo_Node'):
                obj.modifiers['ANM_Geo_Node'].node_group.nodes["Set Material"].inputs[2].default_value = new_mat
            bpy.ops.outliner.orphans_purge()
        elif not obj.active_material:
            Import_ANM_Mat()
            mat = bpy.data.materials['ANM_Mat']
            obj.data.materials.append(mat)
            

        return {'FINISHED'}




class Reset_Children_Objects_OP(Operator):
    """Add relevant bevel profile objects, make object unique after duplicate"""
    bl_label = "Reset_Children_Objects"
    bl_idname = "anm_reset_children_objects.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        if obj:
            if obj.type == 'CURVES' or obj.type == 'CURVE':
                c = True    
        return c

    def execute(self, context):
        hair_cur = context.active_object
        
        animehair_props =  context.scene.animehair_props
        if hair_cur.mode != "OBJECT":
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.transform_apply(location=False, rotation=True, scale=True)

        try:
            if not bpy.data.collections.get('ANM_Curves_profile'):
                Import_curves_presets()

            elif bpy.data.collections.get('ANM_Curves_profile') and animehair_props.bool_source_file_updated:
                ANM_Curves_collection = bpy.data.collections.get('ANM_Curves_profile')
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()
                
                Import_curves_presets()

                animehair_props.bool_source_file_updated = False

            elif len(bpy.data.collections['ANM_Curves_profile'].all_objects) <12:
                ANM_Curves_collection = bpy.data.collections['ANM_Curves_profile']
                for obj in ANM_Curves_collection.objects:
                    bpy.data.objects.remove(obj, do_unlink=True)
                bpy.data.collections.remove(ANM_Curves_collection)
                bpy.ops.outliner.orphans_purge()

                Import_curves_presets()

            has_children = False
            for obj in hair_cur.children:
                if "ANM_Taper" in obj.name:
                    has_children = True
            
            if not has_children:
                cur_coll = bpy.data.collections['ANM_Curves_profile']
                cur_coll.hide_viewport = True
                bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

                hair_cur_coll = hair_cur.users_collection[0]

                cur_t = None
                
                for cur in cur_coll.objects:
                    if cur.name[:5] == 'ANM_T':
                        cur_t = bpy.data.objects.new(cur.name, cur.data)
                        cur_t.data = cur_t.data.copy()
                        hair_cur_coll.objects.link(cur_t)

                        if hair_cur.type == 'CURVE':
                            hair_cur.data.use_path = False
                        cur_t.parent = hair_cur

                        cur_t.hide_viewport = True
                        cur_t.hide_render = True
                for cur in bevel_coll.objects:
                    if cur.type  == 'CURVE':
                        cur_b = bpy.data.objects.new(cur.name, cur.data)
                        cur_b.data = cur_b.data.copy()
                        
                        hair_cur_coll.objects.link(cur_b)

                        cur_t.data.use_path = False
                        cur_b.parent = cur_t

                        cur_b.hide_viewport = True
                        cur_b.hide_render = True

                
            if not hair_cur.modifiers.get('ANM_Geo_Node') and hair_cur.type == 'CURVE':
                taper_obj = None
                for obj in hair_cur.children:
                    if "ANM_Taper" in obj.name:
                        taper_obj = obj
                
                hair_cur.data.taper_object = taper_obj
                hair_cur.data.bevel_object = taper_obj.children[0]
                Apply_existing_taper_value(hair_cur)
            elif hair_cur.modifiers.get('ANM_Geo_Node'):
                taper_obj = None
                for obj in hair_cur.children:
                    if "ANM_Taper" in obj.name:
                        taper_obj = obj
                geo_node = hair_cur.modifiers['ANM_Geo_Node']
                if geo_node.node_group.users > 1:
                    geo_node.node_group = geo_node.node_group.copy()
                    geo_node.node_group.nodes["Object Info"].inputs[0].default_value = taper_obj.children[0]
                     

            bpy.ops.outliner.orphans_purge()

        except:
            bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')
        return {'FINISHED'}



class Convert_Curl_Braids_to_Simple_Hair_OP(Operator):
    """Convert Curl or Braids Hair to Individual Strand Hairs"""
    bl_label = "Convert_Curl_Braids_to_Simple_Hair_OP"
    bl_idname = "anm_convert_curl_braids_to_simple_hair.operator"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        c = False
        obj = context.active_object
        objs = context.selected_objects
        if obj and len(objs) == 1:
            if obj.type == 'CURVE':
                c = True    
        return c

    def execute(self, context):
        hair_cur = context.active_object
        orig_indx = 0

        mat = hair_cur.active_material

        if hair_cur.mode != 'OBJECT':
            bpy.ops.object.mode_set(mode='OBJECT')

        bpy.ops.object.duplicate()
        hair_cur.hide_set(True)
        hair_cur.hide_viewport = True
        hair_cur.hide_render = True

        obj = context.active_object


        if obj.modifiers.get('ANM_Geo_Node'):
            try:
                geo_node = obj.modifiers["ANM_Geo_Node"]

                geo_node.node_group = geo_node.node_group.copy()

                if geo_node.node_group.nodes.get('profile_curve').mapping.curves[0].points[-1].location.y < 0.02:
                    geo_node.node_group.nodes.get('profile_curve').mapping.curves[0].points[-1].location.y = 0.02
                    geo_node.node_group.nodes.get('profile_curve').mapping.update()
                    node_g = geo_node.node_group
                    geo_node.node_group = None
                    geo_node.node_group = node_g
                if geo_node.node_group.nodes.get('profile_curve').mapping.curves[0].points[0].location.y < 0.02:
                    geo_node.node_group.nodes.get('profile_curve').mapping.curves[0].points[0].location.y = 0.02
                    geo_node.node_group.nodes.get('profile_curve').mapping.update()
                    node_g = geo_node.node_group
                    geo_node.node_group = None
                    geo_node.node_group = node_g


                bevel_res = geo_node["Input_4"]
                bevel_rad = geo_node["Input_5"]
                
                if bpy.app.version[0]==4 and bpy.app.version[1]<1 or bpy.app.version[0] <4:
                    use_custom_profile = geo_node.node_group.nodes["Switch"].inputs[1].default_value 
                else:
                    use_custom_profile = geo_node.node_group.nodes["Switch"].inputs[0].default_value 

                cur_t = None
                for ob in obj.children:
                    if "ANM_Taper" in ob.name:
                        cur_t = ob
                        break
            
                if use_custom_profile and cur_t and len(cur_t.children):
                    orig_indx = cur_t.children.index(geo_node.node_group.nodes["Object Info"].inputs[0].default_value)
                   
                geo_node.node_group.nodes["Curve to Mesh"].inputs[2].default_value = False
                geo_node.node_group.nodes["Value"].outputs[0].default_value = 0

                geo_node["Input_5"] = 0.0
                obj.update_tag()


                bpy.ops.object.convert(target='MESH')
                bpy.ops.object.mode_set(mode='EDIT')
                bpy.ops.mesh.select_all(action='SELECT')
                bpy.ops.mesh.remove_doubles(threshold=0.00001)
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.convert(target='CURVE')
                bpy.ops.object.shade_smooth()

                if not cur_t:
                    
                    if not bpy.data.collections.get('ANM_Curves_profile'):
                        Import_curves_presets()
                    
                    cur_coll = bpy.data.collections['ANM_Curves_profile']
                    cur_coll.hide_viewport = True
                    bevel_coll = bpy.data.collections['ANM_Bevel_Profiles']

                    hair_cur_coll = obj.users_collection[0]
                    
                    for cur in cur_coll.objects:
                        if cur.name[:5] == 'ANM_T':
                            cur_t = bpy.data.objects.new(cur.name, cur.data)
                            cur_t.data = cur_t.data.copy()
                            hair_cur_coll.objects.link(cur_t)
                            obj.data.use_path = False
                            cur_t.parent = obj
                            cur_t.hide_viewport = True
                            cur_t.hide_render = True
                    for cur in bevel_coll.objects:
                        if cur.type  == 'CURVE':
                            cur_b = bpy.data.objects.new(cur.name, cur.data)
                            cur_b.data = cur_b.data.copy()
                            hair_cur_coll.objects.link(cur_b)
                            cur_t.data.use_path = False
                            cur_b.parent = cur_t
                            cur_b.hide_viewport = True
                            cur_b.hide_render = True

                if not use_custom_profile:
                    obj.data.bevel_mode = 'ROUND'
                    obj.data.bevel_resolution = bevel_res
                    obj.data.bevel_depth = bevel_rad * 0.75
                else:
                    obj.data.bevel_mode = 'OBJECT'
                    obj.data.bevel_object = cur_t.children[orig_indx]


                if "ANM_Curl_" in obj.name:
                    obj.name = obj.name.replace("ANM_Curl_","ANM_Simple_")
                elif "ANM_Braid2_" in obj.name:
                    obj.name = obj.name.replace("ANM_Braid2_","ANM_Simple_")
                elif "ANM_Braid3_" in obj.name:
                    obj.name = obj.name.replace("ANM_Braid3_","ANM_Simple_")
                elif "ANM_Braid4_" in obj.name:
                    obj.name = obj.name.replace("ANM_Braid4_","ANM_Simple_")
                elif "ANM_Braid5_" in obj.name:
                    obj.name = obj.name.replace("ANM_Braid5_","ANM_Simple_")
                elif "ANM_Braid6_" in obj.name:
                    obj.name = obj.name.replace("ANM_Braid6_","ANM_Simple_")

                obj.data.materials.append(mat)

                bpy.ops.outliner.orphans_purge()

            except:
                bpy.context.window_manager.popup_menu(error_happened, title="Error", icon='ERROR')

        return {'FINISHED'}



class ANM_Preferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    ui_tab_name: bpy.props.StringProperty(default="ANM")
    filepath: bpy.props.StringProperty(name="Bevels Blender File Path",subtype='FILE_PATH',default=str(Path(__file__).parent))

    def draw(self, context):
        
        wm = bpy.context.window_manager
        kc = wm.keyconfigs.user
        ws = bpy.data.workspaces

        layout = self.layout
        layout.prop(self, 'ui_tab_name',text = 'UI_Tab')

        layout.prop(self, 'filepath')
        layout.operator("anm_refresh_bevel_profiles.operator", text='Refresh Bevel Profiles From File Path')

        col = layout.column()

        col.label(text = 'Custom Shortcuts: ')

        old_km_name = ""
        get_kmi_l = []
        for km_add, kmi_add in addon_keymaps:
            for km_con in kc.keymaps:
                if km_add.name == km_con.name:
                    km = km_con
                    break

            for kmi_con in km.keymap_items:
                if kmi_add.idname == kmi_con.idname:
                    if kmi_add.name == kmi_con.name:
                        get_kmi_l.append((km,kmi_con))

        get_kmi_l = sorted(set(get_kmi_l), key=get_kmi_l.index)
        for km, kmi in get_kmi_l:
            if not km.name == old_km_name:
                col.label(text=str(km.name),icon="DOT")
            col.context_pointer_set("keymap", km)
            rna_keymap_ui.draw_kmi([], kc, km, kmi, col, 0)
            col.separator()
            old_km_name = km.name

def update_anim_hair_tab_name():
    interface_classes = (CURVE_TO_ANIME_HAIR_PT_main,)

    for cl in reversed(interface_classes):
        try:
            bpy.utils.unregister_class(cl)
        except:
            pass

    ANIME_HAIR_MAKER.bl_category = bpy.context.preferences.addons[__package__].preferences.ui_tab_name

    for cl in interface_classes:
        bpy.utils.register_class(cl)



addon_keymaps = []

classes = [

    CURVE_TO_ANIME_HAIR_PT_main,  Make_Anime_Hair_OP,  Simplify_Hair_OP, 
    Decimate_Curve_Points_OP, AnimeHairProperties,Previous_Hair_OP,Next_Hair_OP,
    AHM_Select_Main_Handle_OP,AHM_Select_Left_Handle_OP,AHM_Select_Right_Handle_OP,
    AHM_Isolate_Selected_Strand_Show_All_OP,AHM_Hide_Selected_Strand_OP,AHM_Selected_Next_Point_OP,AHM_Selected_Prev_Point_OP,
    Convert_Particle_Hair_To_Curve_OP,Separate_To_New_Curve_OP,
    ANM_Preferences,Add_Curve_To_Source_File_OP,Finalize_ANM_Hair_OP,Edit_ANM_Hair_Bevel_Object_OP,Hide_ANM_Hair_Bevel_Objects_OP,
    Create_Bones_From_Curves_OP,Add_ANM_Hair_On_Target_OP,Draw_Bezier_Hairs_OP,Move_Curve_Point_OP,Refresh_Bevel_Profiles_OP,
    Anime_Hair_With_New_Hair_System_OP,Make_Braids_Hair_OP,Reset_Hair_Curve_Tilt_OP,Reset_Hair_Curve_Radius_OP,Make_Curl_Hair_OP,
    Finalize_Curl_Braids_Hair_OP,Edit_ANM_Hair_Geo_Node_Profile_Object_OP,Convert_Hair_Curves_to_Simple_Curve_OP,
    Convert_Hair_Curves_to_Mesh_OP,Reset_Children_Objects_OP,Convert_Curl_Braids_to_Simple_Hair_OP,Make_ANM_Mat_Unique_OP,
]


def register():
    for cls in classes:
        bpy.utils.register_class(cls)

    update_anim_hair_tab_name()
    
    bpy.types.Scene.animehair_props = bpy.props.PointerProperty(
        type=AnimeHairProperties)

    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon

    if kc:
        km = kc.keymaps.new(name='3D View', space_type='VIEW_3D')

        kmi = km.keymap_items.new("anm_move_curve_point.operator", type='D', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_draw_bezier_hairs.operator", type='NONE', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))
        
        kmi = km.keymap_items.new("anm_add_hair_on_target.operator", type='NONE', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))
        
        kmi = km.keymap_items.new("anime_hair_with_new_hair_system.operator", type='NONE', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_select_main_handle.operator", type='TWO', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_select_left_handle.operator", type='ONE', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_select_right_handle.operator", type='THREE', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_isolate_selected_strand_show_all.operator", type='ACCENT_GRAVE', value='PRESS')
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_hide_selected_strand.operator", type='ACCENT_GRAVE', value='PRESS',alt=True)
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_separate_to_new_curve.operator", type='ACCENT_GRAVE', value='PRESS',ctrl=True,shift=True)
        kmi.active = False
        addon_keymaps.append((km, kmi))
        
        kmi = km.keymap_items.new("anm_selected_prev_point.operator", type='WHEELUPMOUSE', value='PRESS',ctrl=True)
        kmi.active = False
        addon_keymaps.append((km, kmi))

        kmi = km.keymap_items.new("anm_selected_next_point.operator", type='WHEELDOWNMOUSE', value='PRESS',ctrl=True)
        kmi.active = False
        addon_keymaps.append((km, kmi))
    bpy.app.translations.register(__package__, GetTranslationDict())


def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)

    del bpy.types.Scene.animehair_props
    
    for km, kmi in addon_keymaps:
        km.keymap_items.remove(kmi)
    addon_keymaps.clear()
    bpy.app.translations.unregister(__package__)

if __name__ == "__main__":
    register()
